#!/usr/bin/env python3

import os

from datetime import datetime, timezone

timestamp = os.environ.get('SOURCE_DATE_EPOCH', None)
timeformat = '%c %Z'

if timestamp:
    print(datetime.fromtimestamp(int(timestamp), tz=timezone.utc).strftime(timeformat))
else:
    print(datetime.now(tz=timezone.utc).strftime(timeformat))



#!/usr/bin/env python3
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#
#
# Regroup array values into smaller groups separated by numbered comments.
#
# If --check is the first parameter, the script will return
# a non-zero value if a file is not grouped correctly.
# Otherwise the files are regrouped in place.

import argparse
import re
import subprocess
import sys


def regroup_caps(check, filename, start_regex, end_regex,
                 trailing_newline, counter_prefix):
    step = 5

    original = []
    with open(filename, "r") as fh:
        for line in fh:
            original.append(line)

    fixed = []
    game_on = False
    counter = 0
    for line in original:
        line = line.rstrip("\n")
        if game_on:
            if re.search(r'''.*/\* [0-9]+ \*/.*''', line):
                continue
            if re.search(r'''^\s*$''', line):
                continue
            if counter % step == 0:
                if counter != 0:
                    fixed.append("\n")
                fixed.append("%s/* %d */\n" % (counter_prefix, counter))

            if not (line.find("/*") != -1 and line.find("*/") == -1):
                # count two-line comments as one line
                counter = counter + 1

        if re.search(start_regex, line):
            game_on = True
        elif game_on and re.search(end_regex, line):
            if (counter - 1) % step == 0:
                fixed = fixed[:-1]  # /* $counter */
                if counter != 1:
                    fixed = fixed[:-1]  # \n

            if trailing_newline:
                fixed.append("\n")

            game_on = False

        fixed.append(line + "\n")

    if check:
        orig = "".join(original)
        new = "".join(fixed)
        if new != orig:
            diff = subprocess.Popen(["diff", "-u", filename, "-"],
                                    stdin=subprocess.PIPE)
            diff.communicate(input=new.encode('utf-8'))

            print("Incorrect line wrapping in $file",
                  file=sys.stderr)
            print("Use group-qemu-caps.py to generate data files",
                  file=sys.stderr)
            return False
    else:
        with open(filename, "w") as fh:
            for line in fixed:
                print(line, file=fh, end='')

    return True


parser = argparse.ArgumentParser(description='QEMU capabilities group formatter')
parser.add_argument('--check', action="store_true",
                    help='check existing files only')
parser.add_argument('--prefix', default='',
                    help='source code tree prefix')
args = parser.parse_args()

errs = False

if not regroup_caps(args.check,
                    args.prefix + 'src/qemu/qemu_capabilities.c',
                    r'virQEMUCaps grouping marker',
                    r'\);',
                    0,
                    "              "):
    errs = True

if not regroup_caps(args.check,
                    args.prefix + 'src/qemu/qemu_capabilities.h',
                    r'virQEMUCapsFlags grouping marker',
                    r'QEMU_CAPS_LAST \/\* this must',
                    1,
                    "    "):
    errs = True

if errs:
    sys.exit(1)
sys.exit(0)



#!/usr/bin/env python3

#
# esx_vi_generator.py: generates most of the SOAP type mapping code
#
# Copyright (C) 2014 Red Hat, Inc.
# Copyright (C) 2010-2012 Matthias Bolte <matthias.bolte@googlemail.com>
# Copyright (C) 2013 Ata E Husain Bohra <ata.husain@hotmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#

import sys
import os
import os.path


OCCURRENCE__REQUIRED_ITEM = "r"
OCCURRENCE__REQUIRED_LIST = "rl"
OCCURRENCE__OPTIONAL_ITEM = "o"
OCCURRENCE__OPTIONAL_LIST = "ol"
OCCURRENCE__IGNORED = "i"

valid_occurrences = [OCCURRENCE__REQUIRED_ITEM,
                     OCCURRENCE__REQUIRED_LIST,
                     OCCURRENCE__OPTIONAL_ITEM,
                     OCCURRENCE__OPTIONAL_LIST,
                     OCCURRENCE__IGNORED]

autobind_names = set()

separator = "/* " + ("* " * 37) + "*\n"


def aligned(left, right, length=59):
    return left.ljust(length, ' ') + right


class Member:
    def __init__(self, type, occurrence):
        self.type = type
        self.occurrence = occurrence

    def is_enum(self):
        return self.type in predefined_enums or self.type in enums_by_name

    def is_object(self):
        return self.type in predefined_objects or self.type in objects_by_name

    def is_type_generated(self):
        return self.type in enums_by_name or self.type in objects_by_name

    def get_occurrence_comment(self):
        occurrence_map = {
            OCCURRENCE__REQUIRED_ITEM: "/* required */",
            OCCURRENCE__REQUIRED_LIST: "/* required, list */",
            OCCURRENCE__OPTIONAL_ITEM: "/* optional */",
            OCCURRENCE__OPTIONAL_LIST: "/* optional, list */"
        }
        try:
            return occurrence_map[self.occurrence]
        except KeyError:
            raise ValueError("unknown occurrence value '%s'" % self.occurrence)


class Parameter(Member):
    def __init__(self, type, name, occurrence):
        Member.__init__(self, type, occurrence)

        if ':' in name and name.startswith("_this"):
            self.name, self.autobind_name = name.split(":")
        else:
            self.name = name
            self.autobind_name = None

    def generate_parameter(self, is_last=False, is_header=True, offset=0):
        if self.occurrence == OCCURRENCE__IGNORED:
            raise ValueError("invalid function parameter occurrence value '%s'"
                             % self.occurrence)
        elif self.autobind_name is not None:
            return ""
        else:
            string = "       "
            string += " " * offset
            string += "%s%s" % (self.get_type_string(), self.name)

            if is_last:
                if is_header:
                    string += "); "
                else:
                    string += "), "
            else:
                string += ", "

            return aligned(string, self.get_occurrence_comment() + "\n")

    def generate_return(self, offset=0, end_of_line=";"):
        if self.occurrence == OCCURRENCE__IGNORED:
            raise ValueError("invalid function parameter occurrence value '%s'"
                             % self.occurrence)
        else:
            string = "       "
            string += " " * offset
            string += "%s%s)%s" \
                      % (self.get_type_string(True), self.name, end_of_line)

            return aligned(string, self.get_occurrence_comment() + "\n")

    def generate_require_code(self):
        if self.occurrence in [OCCURRENCE__REQUIRED_ITEM,
                               OCCURRENCE__REQUIRED_LIST]:
            return "    ESX_VI__METHOD__PARAMETER__REQUIRE(%s)\n" % self.name
        else:
            return ""

    def generate_serialize_code(self):
        if self.occurrence in [OCCURRENCE__REQUIRED_LIST,
                               OCCURRENCE__OPTIONAL_LIST]:
            return "    ESX_VI__METHOD__PARAMETER__SERIALIZE_LIST(%s, %s)\n" \
                   % (self.type, self.name)
        elif self.type == "String":
            return "    ESX_VI__METHOD__PARAMETER__SERIALIZE_VALUE(String, %s)\n" \
                   % self.name
        else:
            return "    ESX_VI__METHOD__PARAMETER__SERIALIZE(%s, %s)\n" \
                   % (self.type, self.name)

    def get_type_string(self, as_return_value=False):
        string = ""

        if self.type == "String" and \
           self.occurrence not in [OCCURRENCE__REQUIRED_LIST,
                                   OCCURRENCE__OPTIONAL_LIST]:
            if as_return_value:
                string += "char *"
            else:
                string += "const char *"
        elif self.is_enum():
            string += "esxVI_%s " % self.type
        else:
            string += "esxVI_%s *" % self.type

        if as_return_value:
            string += "*"

        return string

    def get_occurrence_short_enum(self):
        if self.occurrence == OCCURRENCE__REQUIRED_ITEM:
            return "RequiredItem"
        elif self.occurrence == OCCURRENCE__REQUIRED_LIST:
            return "RequiredList"
        elif self.occurrence == OCCURRENCE__OPTIONAL_ITEM:
            return "OptionalItem"
        elif self.occurrence == OCCURRENCE__OPTIONAL_LIST:
            return "OptionalList"

        raise ValueError("unknown occurrence value '%s'" % self.occurrence)


class Method:
    def __init__(self, name, parameters, returns):
        self.name = name
        self.parameters = []
        self.autobind_parameter = None
        self.returns = returns

        for parameter in parameters:
            if parameter.autobind_name is None:
                self.parameters.append(parameter)
            else:
                self.autobind_parameter = parameter

    def generate_header(self):
        header = "int esxVI_%s\n" % self.name
        header += "      (esxVI_Context *ctx"

        if len(self.parameters) > 0 or self.returns is not None:
            header += ",\n"

            for parameter in self.parameters[:-1]:
                header += parameter.generate_parameter()

            if self.returns is None:
                header += self.parameters[-1].generate_parameter(is_last=True)
            else:
                header += self.parameters[-1].generate_parameter()
                header += self.returns.generate_return()
        else:
            header += ");\n"

        header += "\n"

        return header

    def generate_source(self):
        source = "/* esxVI_%s */\n" % self.name
        source += "ESX_VI__METHOD(%s," % self.name

        if self.autobind_parameter is not None:
            autobind_names.add(self.autobind_parameter.autobind_name)
            source += " %s,\n" % self.autobind_parameter.autobind_name
        else:
            source += " /* explicit _this */,\n"

        source += "               (esxVI_Context *ctx"

        if len(self.parameters) > 0 or self.returns is not None:
            source += ",\n"

            for parameter in self.parameters[:-1]:
                source += parameter.generate_parameter(is_header=False,
                                                       offset=9)

            if self.returns is None:
                source += self.parameters[-1].generate_parameter(is_last=True,
                                                                 is_header=False,
                                                                 offset=9)
            else:
                source += self.parameters[-1].generate_parameter(is_header=False,
                                                                 offset=9)
                source += self.returns.generate_return(offset=9,
                                                       end_of_line=",")
        else:
            source += "),\n"

        if self.returns is None:
            source += "               void, /* nothing */, None,\n"
        elif self.returns.type == "String":
            source += "               String, Value, %s,\n" \
                      % self.returns.get_occurrence_short_enum()
        else:
            source += "               %s, /* nothing */, %s,\n" \
                      % (self.returns.type,
                         self.returns.get_occurrence_short_enum())

        source += "{\n"

        if self.autobind_parameter is not None:
            source += self.autobind_parameter.generate_require_code()

        for parameter in self.parameters:
            source += parameter.generate_require_code()

        source += "},\n"
        source += "{\n"

        if self.autobind_parameter is not None:
            source += self.autobind_parameter.generate_serialize_code()

        for parameter in self.parameters:
            source += parameter.generate_serialize_code()

        source += "})\n\n\n\n"

        return source


class Property(Member):
    def __init__(self, type, name, occurrence):
        Member.__init__(self, type, occurrence)

        self.name = name

    def generate_struct_member(self):
        if self.occurrence == OCCURRENCE__IGNORED:
            return "    /* FIXME: %s is currently ignored */\n" % self.name
        else:
            string = "    %s%s; " % (self.get_type_string(), self.name)

            return aligned(string, self.get_occurrence_comment() + "\n")

    def generate_free_code(self):
        if self.type == "String" and \
           self.occurrence not in [OCCURRENCE__REQUIRED_LIST,
                                   OCCURRENCE__OPTIONAL_LIST,
                                   OCCURRENCE__IGNORED]:
            return "    VIR_FREE(item->%s);\n" % self.name
        elif self.is_enum():
            return ""
        else:
            if self.occurrence == OCCURRENCE__IGNORED:
                return "    /* FIXME: %s is currently ignored */\n" % self.name
            else:
                return "    esxVI_%s_Free(&item->%s);\n" % (self.type, self.name)

    def generate_validate_code(self, managed=False):
        if managed:
            macro = "ESX_VI__TEMPLATE__PROPERTY__MANAGED_REQUIRE"
        else:
            macro = "ESX_VI__TEMPLATE__PROPERTY__REQUIRE"

        if self.occurrence in [OCCURRENCE__REQUIRED_ITEM,
                               OCCURRENCE__REQUIRED_LIST]:
            return "    %s(%s)\n" % (macro, self.name)
        elif self.occurrence == OCCURRENCE__IGNORED:
            return "    /* FIXME: %s is currently ignored */\n" % self.name
        else:
            return ""

    def generate_deep_copy_code(self):
        if self.occurrence == OCCURRENCE__IGNORED:
            return "    /* FIXME: %s is currently ignored */\n" % self.name
        elif self.occurrence in [OCCURRENCE__REQUIRED_LIST,
                                 OCCURRENCE__OPTIONAL_LIST]:
            return "    ESX_VI__TEMPLATE__PROPERTY__DEEP_COPY_LIST(%s, %s)\n" \
                   % (self.type, self.name)
        elif self.type == "String":
            return "    ESX_VI__TEMPLATE__PROPERTY__DEEP_COPY_VALUE(String, %s)\n" \
                   % self.name
        elif self.is_enum():
            return "    (*dest)->%s = src->%s;\n" % (self.name, self.name)
        else:
            return "    ESX_VI__TEMPLATE__PROPERTY__DEEP_COPY(%s, %s)\n" \
                   % (self.type, self.name)

    def generate_serialize_code(self):
        if self.occurrence == OCCURRENCE__IGNORED:
            return "    /* FIXME: %s is currently ignored */\n" % self.name
        elif self.occurrence in [OCCURRENCE__REQUIRED_LIST,
                                 OCCURRENCE__OPTIONAL_LIST]:
            return "    ESX_VI__TEMPLATE__PROPERTY__SERIALIZE_LIST(%s, %s)\n" \
                   % (self.type, self.name)
        elif self.type == "String":
            return "    ESX_VI__TEMPLATE__PROPERTY__SERIALIZE_VALUE(String, %s)\n" \
                   % self.name
        else:
            return "    ESX_VI__TEMPLATE__PROPERTY__SERIALIZE(%s, %s)\n" \
                   % (self.type, self.name)

    def generate_deserialize_code(self):
        if self.occurrence == OCCURRENCE__IGNORED:
            return "    ESX_VI__TEMPLATE__PROPERTY__DESERIALIZE_IGNORE(%s) /* FIXME */\n" \
                   % self.name
        elif self.occurrence in [OCCURRENCE__REQUIRED_LIST,
                                 OCCURRENCE__OPTIONAL_LIST]:
            return "    ESX_VI__TEMPLATE__PROPERTY__DESERIALIZE_LIST(%s, %s)\n" \
                   % (self.type, self.name)
        elif self.type == "String":
            return "    ESX_VI__TEMPLATE__PROPERTY__DESERIALIZE_VALUE(String, %s)\n" \
                   % self.name
        else:
            return "    ESX_VI__TEMPLATE__PROPERTY__DESERIALIZE(%s, %s)\n" \
                   % (self.type, self.name)

    def generate_lookup_code(self):
        if self.occurrence == OCCURRENCE__IGNORED:
            return "    ESX_VI__TEMPLATE__PROPERTY__CAST_FROM_ANY_TYPE_IGNORE(%s) /* FIXME */\n" \
                   % self.name
        elif self.occurrence in [OCCURRENCE__REQUIRED_LIST,
                                 OCCURRENCE__OPTIONAL_LIST]:
            return "    ESX_VI__TEMPLATE__PROPERTY__CAST_LIST_FROM_ANY_TYPE(%s, %s)\n" \
                   % (self.type, self.name)
        elif self.type == "String":
            return "    ESX_VI__TEMPLATE__PROPERTY__CAST_VALUE_FROM_ANY_TYPE(String, %s)\n" \
                   % self.name
        else:
            return "    ESX_VI__TEMPLATE__PROPERTY__CAST_FROM_ANY_TYPE(%s, %s)\n" \
                   % (self.type, self.name)

    def get_type_string(self):
        if self.type == "String" and \
           self.occurrence not in [OCCURRENCE__REQUIRED_LIST,
                                   OCCURRENCE__OPTIONAL_LIST]:
            return "char *"
        elif self.is_enum():
            return "esxVI_%s " % self.type
        else:
            return "esxVI_%s *" % self.type


class Type:
    def __init__(self, kind, name):
        self.kind = kind
        self.name = name

    def generate_typedef(self):
        return "typedef %s _esxVI_%s esxVI_%s;\n" \
               % (self.kind, self.name, self.name)

    def generate_typeenum(self):
        return "    esxVI_Type_%s,\n" % self.name

    def generate_typetostring(self):
        string = "          case esxVI_Type_%s:\n" % self.name
        string += "            return \"%s\";\n\n" % self.name

        return string

    def generate_typefromstring(self):
        string = "           if (STREQ(type, \"%s\"))\n" % self.name
        string += "               return esxVI_Type_%s;\n" % self.name

        return string


class GenericObject(Type):
    FEATURE__DYNAMIC_CAST = (1 << 1)
    FEATURE__LIST = (1 << 2)
    FEATURE__DEEP_COPY = (1 << 3)
    FEATURE__ANY_TYPE = (1 << 4)
    FEATURE__SERIALIZE = (1 << 5)
    FEATURE__DESERIALIZE = (1 << 6)

    def __init__(self, name, category, managed, generic_objects_by_name):
        Type.__init__(self, "struct", name)
        self.category = category
        self.managed = managed
        self.generic_objects_by_name = generic_objects_by_name

    def generate_comment(self):
        comment = separator
        comment += " * %s: %s\n" % (self.category, self.name)

        if self.extends is not None:
            comment += " * %s  extends %s\n" \
                       % (' ' * len(self.category), self.extends)

        first = True

        if self.extended_by is not None:
            for extended_by in self.extended_by:
                if first:
                    comment += " * %s  extended by %s\n" \
                               % (' ' * len(self.category), extended_by)
                    first = False
                else:
                    comment += " * %s              %s\n" \
                               % (' ' * len(self.category), extended_by)

        comment += " */\n\n"

        return comment

    def generate_struct_members(self, add_banner=False, struct_gap=False):
        members = ""

        if struct_gap:
            members += "\n"

        if self.extends is not None:
            obj = self.generic_objects_by_name[self.extends]
            members += obj.generate_struct_members(add_banner=True,
                                                   struct_gap=False) + "\n"

        if self.extends is not None or add_banner:
            members += "    /* %s */\n" % self.name

        for property in self.properties:
            members += property.generate_struct_member()

        if len(self.properties) < 1:
            members += "    /* no properties */\n"

        return members

    def generate_dispatch(self, suffix, is_first=True):
        source = ""

        if self.extended_by is not None:
            if not is_first:
                source += "\n"

            source += "    /* %s */\n" % self.name

            for extended_by in self.extended_by:
                source += "    ESX_VI__TEMPLATE__DISPATCH__%s(%s)\n" \
                          % (suffix, extended_by)

            for extended_by in self.extended_by:
                obj = self.generic_objects_by_name[extended_by]
                source += obj.generate_dispatch(suffix, False)

        return source

    def generate_free_code(self, add_banner=False):
        source = ""

        if self.extends is not None:
            obj = self.generic_objects_by_name[self.extends]
            source += obj.generate_free_code(add_banner=True) + "\n"

        if self.extends is not None or add_banner:
            source += "    /* %s */\n" % self.name

        if len(self.properties) < 1:
            source += "    /* no properties */\n"
        else:
            string = ""

            for property in self.properties:
                string += property.generate_free_code()

            if len(string) < 1:
                source += "    /* no properties to be freed */\n"
            else:
                source += string

        return source

    def generate_validate_code(self, add_banner=False):
        source = ""

        if self.extends is not None:
            obj = self.generic_objects_by_name[self.extends]
            source += obj.generate_validate_code(add_banner=True) + "\n"

        if self.extends is not None or add_banner:
            source += "    /* %s */\n" % self.name

        if len(self.properties) < 1:
            source += "    /* no properties */\n"
        else:
            string = ""

            for property in self.properties:
                string += property.generate_validate_code(self.managed)

            if len(string) < 1:
                source += "    /* no required properties */\n"
            else:
                source += string

        return source


class Object(GenericObject):
    def __init__(self, name, extends, properties, features=0, extended_by=None):
        GenericObject.__init__(self, name, 'VI Object', False, objects_by_name)
        self.extends = extends
        self.features = features
        self.properties = properties
        self.extended_by = extended_by
        self.candidate_for_dynamic_cast = False

        if self.extended_by is not None:
            self.extended_by.sort()

    def generate_dynamic_cast_code(self, is_first=True):
        source = ""

        if self.extended_by is not None:
            if not is_first:
                source += "\n"

            source += "    /* %s */\n" % self.name

            for extended_by in self.extended_by:
                source += "    ESX_VI__TEMPLATE__DYNAMIC_CAST__ACCEPT(%s)\n" \
                          % extended_by

            for extended_by in self.extended_by:
                obj = objects_by_name[extended_by]
                source += obj.generate_dynamic_cast_code(False)

        return source

    def generate_deep_copy_code(self, add_banner=False):
        source = ""

        if self.extends is not None:
            obj = objects_by_name[self.extends]
            source += obj.generate_deep_copy_code(add_banner=True) + "\n"

        if self.extends is not None or add_banner:
            source += "    /* %s */\n" % self.name

        if len(self.properties) < 1:
            source += "    /* no properties */\n"
        else:
            string = ""

            for property in self.properties:
                string += property.generate_deep_copy_code()

            if len(string) < 1:
                source += "    /* no properties to be deep copied */\n"
            else:
                source += string

        return source

    def generate_serialize_code(self, add_banner=False):
        source = ""

        if self.extends is not None:
            obj = objects_by_name[self.extends]
            source += obj.generate_serialize_code(add_banner=True) + "\n"

        if self.extends is not None or add_banner:
            source += "    /* %s */\n" % self.name

        if len(self.properties) < 1:
            source += "    /* no properties */\n"
        else:
            for property in self.properties:
                source += property.generate_serialize_code()

        return source

    def generate_deserialize_code(self, add_banner=False):
        source = ""

        if self.extends is not None:
            obj = objects_by_name[self.extends]
            source += obj.generate_deserialize_code(add_banner=True) + "\n"

        if self.extends is not None or add_banner:
            source += "    /* %s */\n" % self.name

        if len(self.properties) < 1:
            source += "    /* no properties */\n"
        else:
            for property in self.properties:
                source += property.generate_deserialize_code()

        return source

    def generate_header(self):
        header = self.generate_comment()

        # struct
        header += "struct _esxVI_%s {\n" % self.name

        if self.features & Object.FEATURE__LIST:
            header += aligned("    esxVI_%s *_next; " % self.name,
                              "/* optional */\n")
        else:
            header += aligned("    esxVI_%s *_unused; " % self.name,
                              "/* optional */\n")

        header += aligned("    esxVI_Type _type; ", "/* required */\n")
        header += self.generate_struct_members(struct_gap=True)
        header += "};\n\n"

        # functions
        header += "int esxVI_%s_Alloc(esxVI_%s **item);\n" \
                  % (self.name, self.name)
        header += "void esxVI_%s_Free(esxVI_%s **item);\n" \
                  % (self.name, self.name)
        header += "int esxVI_%s_Validate(esxVI_%s *item);\n" \
                  % (self.name, self.name)

        if self.features & Object.FEATURE__DYNAMIC_CAST:
            if self.extended_by is not None or self.extends is not None:
                header += "esxVI_%s *esxVI_%s_DynamicCast(void *item);\n" \
                          % (self.name, self.name)
            else:
                report_error("cannot add dynamic cast support for an untyped object")

        if self.features & Object.FEATURE__LIST:
            header += "int esxVI_%s_AppendToList(esxVI_%s **list, esxVI_%s *item);\n" \
                      % (self.name, self.name, self.name)

        if self.features & Object.FEATURE__DEEP_COPY:
            header += "int esxVI_%s_DeepCopy(esxVI_%s **dst, esxVI_%s *src);\n" \
                      % (self.name, self.name, self.name)

            if self.features & Object.FEATURE__LIST:
                header += ((
                    "int esxVI_%s_DeepCopyList(esxVI_%s **dstList, "
                    "                          esxVI_%s *srcList);\n") %
                    (self.name, self.name, self.name))

        if self.features & Object.FEATURE__ANY_TYPE:
            header += ((
                "int esxVI_%s_CastFromAnyType(esxVI_AnyType *anyType, "
                "                             esxVI_%s **item);\n") %
                (self.name, self.name))

            if self.features & Object.FEATURE__LIST:
                header += ((
                    "int esxVI_%s_CastListFromAnyType(esxVI_AnyType *anyType, "
                    "                                 esxVI_%s **list);\n") %
                    (self.name, self.name))

        if self.features & Object.FEATURE__SERIALIZE:
            header += ((
                "int esxVI_%s_Serialize(esxVI_%s *item, "
                "                       const char *element, "
                "                       virBufferPtr output);\n") %
                (self.name, self.name))

            if self.features & Object.FEATURE__LIST:
                header += ((
                    "int esxVI_%s_SerializeList(esxVI_%s *list, "
                    "                           const char *element, "
                    "                           virBufferPtr output);\n") %
                    (self.name, self.name))

        if self.features & Object.FEATURE__DESERIALIZE:
            header += "int esxVI_%s_Deserialize(xmlNodePtr node, esxVI_%s **item);\n" \
                      % (self.name, self.name)

            if self.features & Object.FEATURE__LIST:
                header += ((
                    "int esxVI_%s_DeserializeList(xmlNodePtr node, "
                    "                             esxVI_%s **list);\n") %
                    (self.name, self.name))

        header += "\n\n\n"

        return header

    def generate_source(self):
        source = separator
        source += " * VI Object: %s\n" % self.name

        if self.extends is not None:
            source += " *            extends %s\n" % self.extends

        first = True

        if self.extended_by is not None:
            for extended_by in self.extended_by:
                if first:
                    source += " *            extended by %s\n" % extended_by
                    first = False
                else:
                    source += " *                        %s\n" % extended_by

        source += " */\n\n"

        # functions
        source += "/* esxVI_%s_Alloc */\n" % self.name
        source += "ESX_VI__TEMPLATE__ALLOC(%s)\n\n" % self.name

        # free
        source += "/* esxVI_%s_Free */\n" % self.name

        if self.extended_by is None:
            source += "ESX_VI__TEMPLATE__FREE(%s,\n" % self.name
        else:
            source += "ESX_VI__TEMPLATE__DYNAMIC_FREE(%s,\n" % self.name
            source += "{\n"
            source += self.generate_dispatch('FREE')
            source += "},\n"

        source += "{\n"

        if self.features & Object.FEATURE__LIST:
            base_class = get_base_class(self)
            if base_class:
                # avoid "dereferencing type-punned pointer will break
                # strict-aliasing rules" warnings
                source += "    esxVI_%s *baseNext = (esxVI_%s *)item->_next;\n" \
                          % (base_class, base_class)
                source += "    esxVI_%s_Free(&baseNext);\n\n" % base_class
            else:
                source += "    esxVI_%s_Free(&item->_next);\n\n" % self.name

        source += self.generate_free_code()
        source += "})\n\n"

        # validate
        source += "/* esxVI_%s_Validate */\n" % self.name
        source += "ESX_VI__TEMPLATE__VALIDATE(%s,\n" % self.name
        source += "{\n"
        source += self.generate_validate_code()
        source += "})\n\n"

        # dynamic cast
        if self.features & Object.FEATURE__DYNAMIC_CAST:
            if self.extended_by is not None or self.extends is not None:
                source += "/* esxVI_%s_DynamicCast */\n" % self.name
                source += "ESX_VI__TEMPLATE__DYNAMIC_CAST(%s,\n" % self.name
                source += "{\n"
                source += self.generate_dynamic_cast_code()
                source += "})\n\n"
            else:
                report_error("cannot add dynamic cast support for an untyped object")

        # append to list
        if self.features & Object.FEATURE__LIST:
            source += "/* esxVI_%s_AppendToList */\n" % self.name
            source += "ESX_VI__TEMPLATE__LIST__APPEND(%s)\n\n" % self.name

        # deep copy
        if self.features & Object.FEATURE__DEEP_COPY:
            source += "/* esxVI_%s_DeepCopy */\n" % self.name

            if self.extended_by is None:
                source += "ESX_VI__TEMPLATE__DEEP_COPY(%s,\n" % self.name
            else:
                source += "ESX_VI__TEMPLATE__DYNAMIC_DEEP_COPY(%s,\n" % self.name
                source += "{\n"
                source += self.generate_dispatch('DEEP_COPY')
                source += "},\n"

            source += "{\n"
            source += self.generate_deep_copy_code()
            source += "})\n\n"

            if self.features & Object.FEATURE__LIST:
                source += "/* esxVI_%s_DeepCopyList */\n" % self.name
                source += "ESX_VI__TEMPLATE__LIST__DEEP_COPY(%s)\n\n" \
                          % self.name

        # cast from any type
        if self.features & Object.FEATURE__ANY_TYPE:
            source += "/* esxVI_%s_CastFromAnyType */\n" % self.name

            if self.extended_by is None:
                source += "ESX_VI__TEMPLATE__CAST_FROM_ANY_TYPE(%s)\n\n" \
                          % self.name
            else:
                source += "ESX_VI__TEMPLATE__DYNAMIC_CAST_FROM_ANY_TYPE(%s,\n" \
                          % self.name
                source += "{\n"
                source += self.generate_dispatch('CAST_FROM_ANY_TYPE')
                source += "})\n\n"

            if self.features & Object.FEATURE__LIST:
                source += "/* esxVI_%s_CastListFromAnyType */\n" % self.name
                source += "ESX_VI__TEMPLATE__LIST__CAST_FROM_ANY_TYPE(%s)\n\n" \
                          % self.name

        # serialize
        if self.features & Object.FEATURE__SERIALIZE:
            source += "/* esxVI_%s_Serialize */\n" % self.name

            if self.extended_by is None:
                source += "ESX_VI__TEMPLATE__SERIALIZE(%s,\n" % self.name
            else:
                source += "ESX_VI__TEMPLATE__DYNAMIC_SERIALIZE(%s,\n" % self.name
                source += "{\n"
                source += self.generate_dispatch('SERIALIZE')
                source += "},\n"

            source += "{\n"
            source += self.generate_serialize_code()
            source += "})\n\n"

            if self.features & Object.FEATURE__LIST:
                source += "/* esxVI_%s_SerializeList */\n" % self.name
                source += "ESX_VI__TEMPLATE__LIST__SERIALIZE(%s)\n\n" \
                          % self.name

        # deserialize
        if self.features & Object.FEATURE__DESERIALIZE:
            source += "/* esxVI_%s_Deserialize */\n" % self.name

            if self.extended_by is None:
                source += "ESX_VI__TEMPLATE__DESERIALIZE(%s,\n" % self.name
            else:
                source += "ESX_VI__TEMPLATE__DYNAMIC_DESERIALIZE(%s,\n" \
                          % self.name
                source += "{\n"
                source += self.generate_dispatch('DESERIALIZE')
                source += "},\n"

            source += "{\n"
            source += self.generate_deserialize_code()
            source += "})\n\n"

            if self.features & Object.FEATURE__LIST:
                source += "/* esxVI_%s_DeserializeList */\n" % self.name
                source += "ESX_VI__TEMPLATE__LIST__DESERIALIZE(%s)\n\n" \
                          % self.name

        source += "\n\n"

        return source


class ManagedObject(GenericObject):
    def __init__(self, name, extends, properties, features=0, extended_by=None):
        GenericObject.__init__(self, name, 'VI Managed Object', True,
                               managed_objects_by_name)
        self.extends = extends
        self.features = features
        self.properties = properties
        self.extended_by = extended_by

        if self.extended_by is not None:
            self.extended_by.sort()

    def generate_lookup_code1(self, add_banner=False):
        source = ""

        if self.extends is not None:
            obj = managed_objects_by_name[self.extends]
            source += obj.generate_lookup_code1(add_banner=True) + "\n"

        if self.extends is not None or add_banner:
            source += "    /* %s */\n" % self.name

        if len(self.properties) < 1:
            source += "    /* no properties */\n"
        else:
            string = ""

            for property in self.properties:
                string += "    \"%s\\0\"\n" % property.name

            if len(string) < 1:
                source += "    /* no properties */\n"
            else:
                source += string

        return source

    def generate_lookup_code2(self, add_banner=False):
        source = ""

        if self.extends is not None:
            obj = managed_objects_by_name[self.extends]
            source += obj.generate_lookup_code2(add_banner=True) + "\n"

        if self.extends is not None or add_banner:
            source += "    /* %s */\n" % self.name

        if len(self.properties) < 1:
            source += "    /* no properties */\n"
        else:
            string = ""

            for property in self.properties:
                string += property.generate_lookup_code()

            if len(string) < 1:
                source += "    /* no properties */\n"
            else:
                source += string

        return source

    def generate_header(self):
        header = self.generate_comment()

        # struct
        header += "struct _esxVI_%s {\n" % self.name

        if self.features & Object.FEATURE__LIST:
            header += aligned("    esxVI_%s *_next; " % self.name,
                              "/* optional */\n")
        else:
            header += aligned("    esxVI_%s *_unused; " % self.name,
                              "/* optional */\n")

        header += aligned("    esxVI_Type _type; ", "/* required */\n")
        header += aligned("    esxVI_ManagedObjectReference *_reference; ",
                          "/* required */\n")
        header += "\n"
        header += self.generate_struct_members()
        header += "};\n\n"

        # functions
        header += "int esxVI_%s_Alloc(esxVI_%s **item);\n" % (self.name, self.name)
        header += "void esxVI_%s_Free(esxVI_%s **item);\n" % (self.name, self.name)
        header += ("int esxVI_%s_Validate(esxVI_%s *item, "
                   "                      esxVI_String *selectedPropertyNameList);\n") \
            % (self.name, self.name)

        if self.features & Object.FEATURE__LIST:
            header += "int esxVI_%s_AppendToList(esxVI_%s **list, esxVI_%s *item);\n" \
                % (self.name, self.name, self.name)

        header += "\n\n\n"

        return header

    def generate_helper_header(self):
        # functions
        return (
            "int esxVI_Lookup%(name)s(esxVI_Context *ctx,"
            "                         const char *name,"
            "                         esxVI_ManagedObjectReference *root,"
            "                         esxVI_String *selectedPropertyNameList,"
            "                         esxVI_%(name)s **item,"
            "                         esxVI_Occurrence occurrence);\n\n"
            % {"name": self.name}
        )

    def generate_source(self):
        source = self.generate_comment()

        # functions
        source += "/* esxVI_%s_Alloc */\n" % self.name
        source += "ESX_VI__TEMPLATE__ALLOC(%s)\n\n" % self.name

        # free
        source += "/* esxVI_%s_Free */\n" % self.name

        if self.extended_by is None:
            source += "ESX_VI__TEMPLATE__FREE(%s,\n" % self.name
        else:
            source += "ESX_VI__TEMPLATE__DYNAMIC_FREE(%s,\n" % self.name
            source += "{\n"
            source += self.generate_dispatch('FREE')
            source += "},\n"

        source += "{\n"

        if self.features & ManagedObject.FEATURE__LIST:
            if self.extends is not None:
                # avoid "dereferencing type-punned pointer will break
                # strict-aliasing rules" warnings
                source += "    esxVI_%s *next = (esxVI_%s *)item->_next;\n\n" \
                          % (self.extends, self.extends)
                source += "    esxVI_%s_Free(&next);\n" % self.extends
                source += "    item->_next = (esxVI_%s *)next;\n\n" % self.name
            else:
                source += "    esxVI_%s_Free(&item->_next);\n" % self.name

        source += "    esxVI_ManagedObjectReference_Free(&item->_reference);\n\n"
        source += self.generate_free_code()
        source += "})\n\n"

        # validate
        source += "/* esxVI_%s_Validate */\n" % self.name
        source += "ESX_VI__TEMPLATE__MANAGED_VALIDATE(%s,\n" % self.name
        source += "{\n"

        source += self.generate_validate_code()

        source += "})\n\n"

        # append to list
        if self.features & ManagedObject.FEATURE__LIST:
            source += "/* esxVI_%s_AppendToList */\n" % self.name
            source += "ESX_VI__TEMPLATE__LIST__APPEND(%s)\n\n" % self.name

        source += "\n\n"

        return source

    def generate_helper_source(self):
        # lookup
        return (
            "/* esxVI_Lookup%(name)s */\n"
            "ESX_VI__TEMPLATE__LOOKUP(%(name)s,\n"
            "{\n"
            "%(lookup_code1)s},\n"
            "{\n"
            "%(lookup_code2)s})"
            "\n\n\n\n"
            % {"name": self.name,
               "lookup_code1": self.generate_lookup_code1(),
               "lookup_code2": self.generate_lookup_code2()}
        )


class Enum(Type):
    FEATURE__ANY_TYPE = (1 << 1)
    FEATURE__SERIALIZE = (1 << 2)
    FEATURE__DESERIALIZE = (1 << 3)

    def __init__(self, name, values, features=0):
        Type.__init__(self, "enum", name)
        self.values = values
        self.features = features

    def generate_header(self):
        header = separator
        header += " * VI Enum: %s\n" % self.name
        header += " */\n\n"
        header += "enum _esxVI_%s {\n" % self.name
        header += "    esxVI_%s_Undefined = 0,\n" % self.name

        for value in self.values:
            header += "    esxVI_%s_%s,\n" % (self.name, capitalize_first(value))

        header += "};\n\n"

        # functions
        if self.features & Enum.FEATURE__ANY_TYPE:
            header += ("int esxVI_%s_CastFromAnyType(esxVI_AnyType *anyType, "
                       "                             esxVI_%s *item);\n") \
                % (self.name, self.name)

        if self.features & Enum.FEATURE__SERIALIZE:
            header += ("int esxVI_%s_Serialize(esxVI_%s item, const char *element, "
                       "                       virBufferPtr output);\n") \
                % (self.name, self.name)

        if self.features & Enum.FEATURE__DESERIALIZE:
            header += ("int esxVI_%s_Deserialize(xmlNodePtr node, "
                       "                         esxVI_%s *item);\n") \
                % (self.name, self.name)

        header += "\n\n\n"

        return header

    def generate_source(self):
        source = separator
        source += " * VI Enum: %s\n" % self.name
        source += " */\n\n"
        source += "static const esxVI_Enumeration _esxVI_%s_Enumeration = {\n" \
            % self.name
        source += "    esxVI_Type_%s, {\n" % self.name

        for value in self.values:
            source += "        { \"%s\", esxVI_%s_%s },\n" \
                % (value, self.name, capitalize_first(value))

        source += "        { NULL, -1 },\n"
        source += "    },\n"
        source += "};\n\n"

        # functions
        if self.features & Enum.FEATURE__ANY_TYPE:
            source += "/* esxVI_%s_CastFromAnyType */\n" % self.name
            source += "ESX_VI__TEMPLATE__ENUMERATION__CAST_FROM_ANY_TYPE(%s)\n\n" \
                      % self.name

        if self.features & Enum.FEATURE__SERIALIZE:
            source += "/* esxVI_%s_Serialize */\n" % self.name
            source += "ESX_VI__TEMPLATE__ENUMERATION__SERIALIZE(%s)\n\n" \
                      % self.name

        if self.features & Enum.FEATURE__DESERIALIZE:
            source += "/* esxVI_%s_Deserialize */\n" % self.name
            source += "ESX_VI__TEMPLATE__ENUMERATION__DESERIALIZE(%s)\n\n" \
                      % self.name

        source += "\n\n"

        return source


def report_error(message):
    print("error: " + message)
    sys.exit(1)


def capitalize_first(string):
    return string[:1].upper() + string[1:]


def parse_object(block):
    # expected format: [managed] object <name> [extends <name>]
    header_items = block[0][1].split()
    managed = False

    if header_items[0] == "managed":
        managed = True
        del header_items[0]

    if len(header_items) < 2:
        report_error("line %d: invalid block header" % (number))

    assert header_items[0] == "object"

    name = header_items[1]
    extends = None

    if len(header_items) > 2:
        if header_items[2] != "extends":
            report_error("line %d: invalid block header" % (number))
        else:
            extends = header_items[3]

    properties = []

    for line in block[1:]:
        # expected format: <type> <name> <occurrence>
        items = line[1].split()

        if len(items) != 3:
            report_error("line %d: invalid property" % line[0])

        if items[2] not in valid_occurrences:
            report_error("line %d: invalid occurrence" % line[0])

        properties.append(Property(type=items[0], name=items[1],
                                   occurrence=items[2]))

    if managed:
        return ManagedObject(name=name, extends=extends, properties=properties)
    else:
        return Object(name=name, extends=extends, properties=properties)


def parse_enum(block):
    # expected format: enum <name>
    header_items = block[0][1].split()

    if len(header_items) < 2:
        report_error("line %d: invalid block header" % (number))

    assert header_items[0] == "enum"

    name = header_items[1]

    values = []

    for line in block[1:]:
        # expected format: <value>
        values.append(line[1])

    return Enum(name=name, values=values)


def parse_method(block):
    # expected format: method <name> [returns <type> <occurrence>]
    header_items = block[0][1].split()

    if len(header_items) < 2:
        report_error("line %d: invalid block header" % (number))

    assert header_items[0] == "method"

    name = header_items[1]
    returns = None

    if len(header_items) > 2:
        if header_items[2] != "returns":
            report_error("line %d: invalid block header" % (number))
        else:
            returns = Parameter(type=header_items[3], name="output",
                                occurrence=header_items[4])

    parameters = []

    for line in block[1:]:
        # expected format: <type> <name> <occurrence>
        items = line[1].split()

        if len(items) != 3:
            report_error("line %d: invalid property" % line[0])

        if items[2] not in valid_occurrences:
            report_error("line %d: invalid occurrence" % line[0])

        parameters.append(Parameter(type=items[0], name=items[1],
                                    occurrence=items[2]))

    return Method(name=name, parameters=parameters, returns=returns)


def is_known_type(type):
    return (type in predefined_objects or
            type in predefined_enums or
            type in objects_by_name or
            type in managed_objects_by_name or
            type in enums_by_name)


def get_base_class(obj):
    if not obj.extends:
        return None
    base_class = None
    try:
        base_class = base_class_by_name[obj.extends]
    except KeyError:
        parent = objects_by_name[obj.extends]
        base_class = get_base_class(parent)
        if not base_class:
            base_class = parent.name
        base_class_by_name[name] = base_class
    return base_class


def open_file(filename):
    return open(filename, "wt")


predefined_enums = ["Boolean"]

predefined_objects = ["AnyType",
                      "Byte",
                      "Int",
                      "Long",
                      "String",
                      "DateTime",
                      "MethodFault",
                      "ManagedObjectReference"]

additional_enum_features = {
    "ManagedEntityStatus": Enum.FEATURE__ANY_TYPE,
    "TaskInfoState": Enum.FEATURE__ANY_TYPE,
    "VirtualMachinePowerState": Enum.FEATURE__ANY_TYPE
}

additional_object_features = {
    "AutoStartDefaults": Object.FEATURE__ANY_TYPE,
    "AutoStartPowerInfo": Object.FEATURE__ANY_TYPE,
    "DatastoreHostMount": (Object.FEATURE__DEEP_COPY | Object.FEATURE__LIST |
                           Object.FEATURE__ANY_TYPE),
    "DatastoreInfo": Object.FEATURE__ANY_TYPE | Object.FEATURE__DYNAMIC_CAST,
    "GuestNicInfo": Object.FEATURE__LIST | Object.FEATURE__ANY_TYPE,
    "HostConfigManager": Object.FEATURE__ANY_TYPE,
    "HostCpuIdInfo": Object.FEATURE__LIST | Object.FEATURE__ANY_TYPE,
    "HostDatastoreBrowserSearchResults": (Object.FEATURE__LIST |
                                          Object.FEATURE__ANY_TYPE),
    "HostHostBusAdapter": Object.FEATURE__LIST | Object.FEATURE__ANY_TYPE,
    "HostInternetScsiHba": (Object.FEATURE__DYNAMIC_CAST |
                            Object.FEATURE__DEEP_COPY),
    "HostInternetScsiTargetTransport": Object.FEATURE__DYNAMIC_CAST,
    "HostScsiDisk": (Object.FEATURE__LIST | Object.FEATURE__ANY_TYPE |
                     Object.FEATURE__DYNAMIC_CAST),
    "HostScsiTopologyInterface": (Object.FEATURE__LIST |
                                  Object.FEATURE__ANY_TYPE),
    "HostScsiTopologyLun": (Object.FEATURE__ANY_TYPE | Object.FEATURE__LIST |
                            Object.FEATURE__DEEP_COPY),
    "HostScsiTopologyTarget": Object.FEATURE__ANY_TYPE | Object.FEATURE__LIST,
    "HostPortGroup": Object.FEATURE__LIST | Object.FEATURE__ANY_TYPE,
    "HostVirtualSwitch": (Object.FEATURE__DEEP_COPY | Object.FEATURE__LIST |
                          Object.FEATURE__ANY_TYPE),
    "ManagedObjectReference": Object.FEATURE__ANY_TYPE,
    "ObjectContent": Object.FEATURE__DEEP_COPY,
    "PhysicalNic": (Object.FEATURE__DEEP_COPY | Object.FEATURE__LIST |
                    Object.FEATURE__ANY_TYPE),
    "ResourcePoolResourceUsage": Object.FEATURE__ANY_TYPE,
    "ScsiLun": (Object.FEATURE__LIST | Object.FEATURE__ANY_TYPE |
                Object.FEATURE__DEEP_COPY),
    "ScsiLunDurableName": Object.FEATURE__LIST,
    "ServiceContent": Object.FEATURE__DESERIALIZE,
    "SharesInfo": Object.FEATURE__ANY_TYPE,
    "TaskInfo": Object.FEATURE__LIST | Object.FEATURE__ANY_TYPE,
    "UserSession": Object.FEATURE__ANY_TYPE,
    "VirtualMachineQuestionInfo": Object.FEATURE__ANY_TYPE,
    "VirtualMachineSnapshotTree": (Object.FEATURE__DEEP_COPY |
                                   Object.FEATURE__ANY_TYPE),
    "VmEventArgument": Object.FEATURE__DESERIALIZE
}

removed_object_features = {}

if len(sys.argv) != 4:
    report_error("usage: %s srcdir builddir header" % sys.argv[0])

input_filename = os.path.join(sys.argv[1], "esx/esx_vi_generator.input")
output_dirname = os.path.join(sys.argv[2], "esx")
header = sys.argv[3] == "header"


if header:
    types_typedef = open_file(os.path.join(output_dirname, "esx_vi_types.generated.typedef"))
    types_typeenum = open_file(os.path.join(output_dirname, "esx_vi_types.generated.typeenum"))
    types_header = open_file(os.path.join(output_dirname, "esx_vi_types.generated.h"))
    methods_header = open_file(os.path.join(output_dirname, "esx_vi_methods.generated.h"))
    helpers_header = open_file(os.path.join(output_dirname, "esx_vi.generated.h"))
else:
    types_typetostring = open_file(os.path.join(output_dirname, "esx_vi_types.generated.typetostring"))
    types_typefromstring = open_file(os.path.join(output_dirname, "esx_vi_types.generated.typefromstring"))
    types_source = open_file(os.path.join(output_dirname, "esx_vi_types.generated.c"))
    methods_macro = open_file(os.path.join(output_dirname, "esx_vi_methods.generated.macro"))
    methods_source = open_file(os.path.join(output_dirname, "esx_vi_methods.generated.c"))
    helpers_source = open_file(os.path.join(output_dirname, "esx_vi.generated.c"))


number = 0
objects_by_name = {}
managed_objects_by_name = {}
enums_by_name = {}
methods_by_name = {}
block = None
base_class_by_name = {}


# parse input file
for line in open(input_filename, "rt").readlines():
    number += 1

    if "#" in line:
        line = line[:line.index("#")]

    line = line.lstrip().rstrip()

    if len(line) < 1:
        continue

    if line.startswith("object") or line.startswith("managed object") or \
       line.startswith("enum") or line.startswith("method"):
        if block is not None:
            report_error("line %d: nested block found" % (number))
        else:
            block = []

    if block is not None:
        if line == "end":
            if block[0][1].startswith("object"):
                obj = parse_object(block)
                objects_by_name[obj.name] = obj
            elif block[0][1].startswith("managed object"):
                obj = parse_object(block)
                managed_objects_by_name[obj.name] = obj
            elif block[0][1].startswith("enum"):
                enum = parse_enum(block)
                enums_by_name[enum.name] = enum
            else:
                method = parse_method(block)
                methods_by_name[method.name] = method

            block = None
        else:
            block.append((number, line))


for method in methods_by_name.values():
    # method parameter types must be serializable
    for parameter in method.parameters:
        if not parameter.is_type_generated():
            continue

        if parameter.is_enum():
            enums_by_name[parameter.type].features |= Enum.FEATURE__SERIALIZE
        else:
            objects_by_name[parameter.type].features |= Object.FEATURE__SERIALIZE
            objects_by_name[parameter.type].candidate_for_dynamic_cast = True

        # detect list usage
        if parameter.occurrence == OCCURRENCE__REQUIRED_LIST or \
           parameter.occurrence == OCCURRENCE__OPTIONAL_LIST:
            if parameter.is_enum():
                report_error("unsupported usage of enum '%s' as list in '%s'"
                             % (parameter.type, method.name))
            else:
                objects_by_name[parameter.type].features |= Object.FEATURE__LIST

    # method return types must be deserializable
    if method.returns and method.returns.is_type_generated():
        if method.returns.is_enum():
            enums_by_name[method.returns.type].features |= Enum.FEATURE__DESERIALIZE
        else:
            objects_by_name[method.returns.type].features |= Object.FEATURE__DESERIALIZE
            objects_by_name[method.returns.type].candidate_for_dynamic_cast = True

        # detect list usage
        if method.returns.occurrence == OCCURRENCE__REQUIRED_LIST or \
           method.returns.occurrence == OCCURRENCE__OPTIONAL_LIST:
            if method.returns.is_enum():
                report_error("unsupported usage of enum '%s' as list in '%s'"
                             % (method.returns.type, method.name))
            else:
                objects_by_name[method.returns.type].features |= Object.FEATURE__LIST


for enum in enums_by_name.values():
    # apply additional features
    if enum.name in additional_enum_features:
        enum.features |= additional_enum_features[enum.name]

        if additional_enum_features[enum.name] & Enum.FEATURE__ANY_TYPE:
            enum.features |= Enum.FEATURE__DESERIALIZE


for obj in objects_by_name.values():
    for property in obj.properties:
        if property.occurrence != OCCURRENCE__IGNORED and \
           not is_known_type(property.type):
            report_error("object '%s' contains unknown property type '%s'"
                         % (obj.name, property.type))

    if obj.extends is not None:
        if not is_known_type(obj.extends):
            report_error("object '%s' extends unknown object '%s'"
                         % (obj.name, obj.extends))

    for property in obj.properties:
        if not property.is_type_generated():
            continue

        if property.is_enum():
            enums_by_name[property.type].candidate_for_dynamic_cast = True
        else:
            objects_by_name[property.type].candidate_for_dynamic_cast = True

        # detect list usage
        if property.occurrence == OCCURRENCE__REQUIRED_LIST or \
           property.occurrence == OCCURRENCE__OPTIONAL_LIST:
            if property.is_enum():
                report_error("unsupported usage of enum '%s' as list in '%s'"
                             % (property.type, obj.type))
            else:
                objects_by_name[property.type].features |= Object.FEATURE__LIST

    # apply/remove additional features
    if obj.name in additional_object_features:
        obj.features |= additional_object_features[obj.name]

        if additional_object_features[obj.name] & Object.FEATURE__ANY_TYPE:
            obj.features |= Object.FEATURE__DESERIALIZE

    if obj.name in removed_object_features:
        obj.features &= ~removed_object_features[obj.name]

    # detect extended_by relation
    if obj.extends is not None:
        extended_obj = objects_by_name[obj.extends]

        if extended_obj.extended_by is None:
            extended_obj.extended_by = [obj.name]
        else:
            extended_obj.extended_by.append(obj.name)
            extended_obj.extended_by.sort()


for obj in objects_by_name.values():
    # if an object is a candidate (it is used directly as parameter or return
    # type or is a member of another object) and it is extended by another
    # object then this type needs the dynamic cast feature
    if obj.candidate_for_dynamic_cast and obj.extended_by:
        obj.features |= Object.FEATURE__DYNAMIC_CAST


def propagate_feature(obj, feature):
    global features_have_changed

    if not (obj.features & feature):
        return

    for property in obj.properties:
        if (property.occurrence == OCCURRENCE__IGNORED or
                not property.is_type_generated()):
            continue

        if property.is_enum():
            if (feature == Object.FEATURE__SERIALIZE and
                not (enums_by_name[property.type].features &
                     Enum.FEATURE__SERIALIZE)):
                enums_by_name[property.type].features |= Enum.FEATURE__SERIALIZE
                features_have_changed = True
            elif (feature == Object.FEATURE__DESERIALIZE and
                  not (enums_by_name[property.type].features &
                       Enum.FEATURE__DESERIALIZE)):
                enums_by_name[property.type].features |= Enum.FEATURE__DESERIALIZE
                features_have_changed = True
        elif property.is_object():
            if not (objects_by_name[property.type].features & feature):
                objects_by_name[property.type].features |= feature
                features_have_changed = True

            if obj.name != property.type:
                propagate_feature(objects_by_name[property.type], feature)


def inherit_features(obj):
    global features_have_changed

    if obj.extended_by is not None:
        for extended_by in obj.extended_by:
            previous = objects_by_name[extended_by].features
            objects_by_name[extended_by].features |= obj.features

            if objects_by_name[extended_by].features != previous:
                features_have_changed = True

    if obj.extends is not None:
        previous = objects_by_name[obj.extends].features
        objects_by_name[obj.extends].features |= obj.features

        if objects_by_name[obj.extends].features != previous:
            features_have_changed = True

    if obj.extended_by is not None:
        for extended_by in obj.extended_by:
            inherit_features(objects_by_name[extended_by])


# there are two directions to spread features:
# 1) up and down the inheritance chain
# 2) from object types to their member property types
# spreading needs to be done alternating on both directions because they can
# affect each other
features_have_changed = True

while features_have_changed:
    features_have_changed = False

    for obj in objects_by_name.values():
        propagate_feature(obj, Object.FEATURE__DEEP_COPY)
        propagate_feature(obj, Object.FEATURE__SERIALIZE)
        propagate_feature(obj, Object.FEATURE__DESERIALIZE)

    for obj in objects_by_name.values():
        inherit_features(obj)


for obj in managed_objects_by_name.values():
    for property in obj.properties:
        if property.occurrence != OCCURRENCE__IGNORED and \
           not is_known_type(property.type):
            report_error("object '%s' contains unknown property type '%s'"
                         % (obj.name, property.type))

    if obj.extends is not None:
        if not is_known_type(obj.extends):
            report_error("object '%s' extends unknown object '%s'"
                         % (obj.name, obj.extends))

    # detect extended_by relation
    if obj.extends is not None:
        extended_obj = managed_objects_by_name[obj.extends]

        if extended_obj.extended_by is None:
            extended_obj.extended_by = [obj.name]
        else:
            extended_obj.extended_by.append(obj.name)
            extended_obj.extended_by.sort()


notice = "/* Generated by esx_vi_generator.py */\n\n\n\n"

if (header):
    types_typedef.write(notice)
    types_typeenum.write(notice)
    types_header.write(notice)
    methods_header.write(notice)
    helpers_header.write(notice)
else:
    types_typetostring.write(notice)
    types_typefromstring.write(notice)
    types_source.write(notice)
    methods_macro.write(notice)
    methods_source.write(notice)
    helpers_source.write(notice)


# output enums
if header:
    types_typedef.write(separator +
                        " * VI Enums\n" +
                        " */\n\n")

names = sorted(enums_by_name.keys())

for name in names:
    if header:
        types_typedef.write(enums_by_name[name].generate_typedef())
        types_typeenum.write(enums_by_name[name].generate_typeenum())
        types_header.write(enums_by_name[name].generate_header())
    else:
        types_typetostring.write(enums_by_name[name].generate_typetostring())
        types_typefromstring.write(enums_by_name[name].generate_typefromstring())
        types_source.write(enums_by_name[name].generate_source())


# output objects
if header:
    types_typedef.write("\n\n\n" +
                        separator +
                        " * VI Objects\n" +
                        " */\n\n")
    types_typeenum.write("\n")
else:
    types_typetostring.write("\n")
    types_typefromstring.write("\n")

names = sorted(objects_by_name.keys())

for name in names:
    if header:
        types_typedef.write(objects_by_name[name].generate_typedef())
        types_typeenum.write(objects_by_name[name].generate_typeenum())
        types_header.write(objects_by_name[name].generate_header())
    else:
        types_typetostring.write(objects_by_name[name].generate_typetostring())
        types_typefromstring.write(objects_by_name[name].generate_typefromstring())
        types_source.write(objects_by_name[name].generate_source())


# output managed objects
if header:
    types_typedef.write("\n\n\n" +
                        separator +
                        " * VI Managed Objects\n" +
                        " */\n\n")
    types_typeenum.write("\n")
else:
    types_typetostring.write("\n")
    types_typefromstring.write("\n")

names = sorted(managed_objects_by_name.keys())

for name in names:
    if header:
        types_typedef.write(managed_objects_by_name[name].generate_typedef())
        types_typeenum.write(managed_objects_by_name[name].generate_typeenum())
        types_header.write(managed_objects_by_name[name].generate_header())
    else:
        types_typetostring.write(managed_objects_by_name[name].generate_typetostring())
        types_typefromstring.write(managed_objects_by_name[name].generate_typefromstring())
        types_source.write(managed_objects_by_name[name].generate_source())


# output methods
names = sorted(methods_by_name.keys())

for name in names:
    if header:
        methods_header.write(methods_by_name[name].generate_header())
    else:
        methods_source.write(methods_by_name[name].generate_source())

if not header:
    names = list(autobind_names)
    names.sort()

    for name in names:
        string = aligned("#define ESX_VI__METHOD__PARAMETER__THIS__%s " % name, "\\\n", 78)
        string += "    ESX_VI__METHOD__PARAMETER__THIS_FROM_SERVICE(ManagedObjectReference,      \\\n"
        string += aligned("", "%s)\n\n\n\n" % name, 49)

        methods_macro.write(string)


# output helpers
names = sorted(managed_objects_by_name.keys())

for name in names:
    if header:
        helpers_header.write(managed_objects_by_name[name].generate_helper_header())
    else:
        helpers_source.write(managed_objects_by_name[name].generate_helper_source())



#!/usr/bin/env python3

import os
import sys

meson_build_root = sys.argv[1]
file_name = sys.argv[2]

meson_dist_root = os.environ['MESON_DIST_ROOT']

os.system('cp {0} {1}'.format(
    os.path.join(meson_build_root, file_name),
    os.path.join(meson_dist_root, file_name)
))



#!/usr/bin/env python3
#
# Copyright (C) 2019 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#
# This uses pdwtags to check remote protocol defs
#
# * the "split" splits on the /* DD */ comments, so that $p iterates
#     through the struct definitions.
# * process only "struct remote_..." entries
# * remove comments and preceding TAB throughout
# * remove empty lines throughout
# * remove white space at end of buffer

import os
import os.path
import re
import subprocess
import sys

name = sys.argv[1]
targetname = sys.argv[2]
libpath = sys.argv[3]
pdwtags = sys.argv[4]
expected = sys.argv[5]

builddir = os.path.dirname(libpath)
libname = os.path.basename(libpath)


def get_subdir(dirname, subdir):
    objectdir = ""
    reg = re.compile(subdir)
    for d in os.listdir(path=dirname):
        if reg.match(d):
            objectdir = d
            break

    if objectdir == "":
        raise Exception("Failed to find '{0}' in '{1}'".format(subdir, dirname))

    return os.path.join(dirname, objectdir)


# Figure out where is the meson target private directory that contains
# generated object files.
# With meson version < 0.55.0 the directory pattern is:
#
#   `hash_string@@target_name@bin_type` for example `25a6634@@vir_net_rpc@sta`
#
# but this was changed in meson 0.55.0 to a new pattern:
#
#   `output_file_name.p` for example `libvirt_net_rpc.a.p`
objectdir = get_subdir(
    builddir,
    r'(.*@{0}@.*|{1}\.p)'.format(targetname, re.escape(libname)))

proto_o = get_subdir(objectdir, r'.*{0}\.c\.o'.format(name))

pdwtagsproc = subprocess.Popen([pdwtags, "--verbose", proto_o],
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE)
out, err = pdwtagsproc.communicate()
out = out.decode("utf-8")
err = err.decode("utf-8")

if out == "" or pdwtagsproc.returncode != 0:
    if out == "":
        print("WARNING: no output, pdwtags appears broken:", file=sys.stderr)
    else:
        print("WARNING: exit code %d, pdwtags appears broken:" %
              pdwtagsproc.returncode, file=sys.stderr)
    for line in err.strip().split("\n"):
        print("WARNING: %s" % line, file=sys.stderr)
    print("WARNING: skipping the remote protocol test", file=sys.stderr)
    sys.exit(0)

# With pdwtags 1.8, --verbose output includes separators like these:
# /* 93 */
# /* <0> (null):0 */
# with the second line omitted for intrinsic types.
# Whereas with pdwtags 1.3, they look like this:
# /* <2d2> /usr/include/libio.h:180 */
# The alternation of the following regexps matches both cases.
r1 = r'''/\* \d+ \*/'''
r2 = r'''/\* <[0-9a-fA-F]+> \S+:\d+ \*/'''

libs_prefix = "remote_|qemu_|lxc_|admin_"
other_prefix = "keepalive|vir(Net|LockSpace|LXCMonitor)"
struct_prefix = "(" + libs_prefix + "|" + other_prefix + ")"

n = 0
bits = re.split(r'''\n*(?:%s|%s)\n''' % (r1, r2), out)
actual = ["/* -*- c -*- */"]

for bit in bits:
    if re.search(r'''^(struct|enum)\s+''' + struct_prefix, bit):
        bit = re.sub(r'''\t*/\*.*?\*/''', "", bit)
        bit = re.sub(r'''\s+\n''', '''\n''', bit)
        bit = re.sub(r'''\s+$''', "", bit)
        bit = re.sub(r'''\t''', "        ", bit)
        actual.append(bit)
        n = n + 1

if n < 1:
    print("WARNING: No structs/enums matched. Your", file=sys.stderr)
    print("WARNING: pdwtags program is probably too old", file=sys.stderr)
    print("WARNING: skipping the remote protocol test", file=sys.stderr)
    print("WARNING: install dwarves-1.3 or newer", file=sys.stderr)
    sys.exit(8)

diff = subprocess.Popen(["diff", "-b", "-u", expected, "-"], stdin=subprocess.PIPE)
actualstr = "\n".join(actual) + "\n"
# i686 builds on x86_64 host add __attribute__(packed)) to
# the structs. This doesn't matter functionally because we
# know our RPC structs are suitably aligned to not need
# packing, so we can just trim the attribute.
actualstr = re.sub(r'''} __attribute__\(\(__packed__\)\);''', "};", actualstr)

diff.communicate(input=actualstr.encode("utf-8"))

sys.exit(diff.returncode)



#!/usr/bin/env python3
#
# Copyright (C) 2011-2019 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#
#
# Generate a set of systemtap probe definitions corresponding to
# DTrace probe markers in libvirt.so
#
#  python dtrace2systemtap.py probes.d > libvirt_probes.stp
#

import re
import sys

file = None
filelist = []
files = {}

bindir = sys.argv[1]
sbindir = sys.argv[2]
libdir = sys.argv[3]
dtrace = sys.argv[4]

probe = None
args = None

# Read the DTraceprobes definition

with open(dtrace, "r") as fh:
    lineno = 0
    for line in fh:
        lineno = lineno + 1
        line = line.strip()
        if line == "":
            continue
        if "provider " in line and "{" in line:
            continue
        if "};" in line:
            continue

        if line.startswith("#"):
            m = re.search(r'''^\#\s*file:\s*(\S+)$''', line)
            if m is not None:
                file = m.group(1)
                filelist.append(file)
                files[file] = {"prefix": None, "probes": []}
                continue

            m = re.search(r'''^\#\s*prefix:\s*(\S+)$''', line)
            if m is not None:
                files[file]["prefix"] = m.group(1)
                continue

            m = re.search(r'''^\#\s*binary:\s*(\S+)$''', line)
            if m is not None:
                files[file]["binary"] = m.group(1)
                continue

            m = re.search(r'''^\#\s*module:\s*(\S+)$''', line)
            if m is not None:
                files[file]["module"] = m.group(1)

            # ignore unknown comments
        else:
            m = re.search(r'''probe\s+([a-zA-Z0-9_]+)\((.*?)(\);)?$''', line)
            if m is not None:
                probe = m.group(1)
                args = m.group(2)
                if m.group(3) is not None:
                    files[file]["probes"].append([probe, args])
                    probe = None
                    args = None
            elif probe is not None:
                m = re.search(r'''^(.*?)(\);)?$''', line)
                if m is not None:
                    args = args + m.group(1)
                    if m.group(2) is not None:
                        files[file]["probes"].append([probe, args])
                        probe = None
                        args = None
                else:
                    raise Exception("unexpected data %s on line %d" %
                                    (line, lineno))
            else:
                raise Exception("unexpected data %s on line %d" %
                                (line, lineno))

# Write out the SystemTap probes
for file in filelist:
    prefix = files[file]["prefix"]
    probes = files[file]["probes"]

    print("# %s\n" % file)
    for probe in probes:
        name = probe[0]
        args = probe[1]

        pname = name.replace(prefix + "_", "libvirt." + prefix + ".")

        binary = libdir + "/libvirt.so"
        if "binary" in files[file]:
            binary = sbindir + "/" + files[file]["binary"]
        if "module" in files[file]:
            binary = libdir + "/" + files[file]["module"]

        print("probe %s = process(\"%s\").mark(\"%s\") {" %
              (pname, binary, name))

        argbits = args.split(",")
        for idx in range(len(argbits)):
            arg = argbits[idx]
            isstr = False
            if 'char *' in arg:
                isstr = True

            m = re.search(r'''^.*\s\*?(\S+)$''', arg)
            if m is not None:
                arg = m.group(1)
            else:
                raise Exception("Malformed arg %s" % arg)

            if isstr:
                print("  %s = user_string($arg%d);" % (arg, idx + 1))
            else:
                print("  %s = $arg%d;" % (arg, idx + 1))
        print("}\n")
    print("")



#!/usr/bin/env python3
#
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#

import re
import sys

drvfiles = []
symfiles = []
for arg in sys.argv:
    if arg.endswith(".h"):
        drvfiles.append(arg)
    else:
        symfiles.append(arg)

symbols = {}

for symfile in symfiles:
    with open(symfile, "r") as fh:
        for line in fh:
            m = re.search(r'''^\s*(vir\w+)\s*;\s*$''', line)
            if m is not None:
                symbols[m.group(1)] = True

status = 0
for drvfile in drvfiles:
    with open(drvfile, "r") as fh:
        for line in fh:
            m = re.search(r'''\*(virDrv\w+)\s*\)''', line)
            if m is not None:
                drv = m.group(1)

                skip = [
                    "virDrvStateInitialize",
                    "virDrvStateCleanup",
                    "virDrvStateReload",
                    "virDrvStateStop",
                    "virDrvStateShutdownPrepare",
                    "virDrvStateShutdownWait",
                    "virDrvConnectSupportsFeature",
                    "virDrvConnectURIProbe",
                    "virDrvDomainMigratePrepare",
                    "virDrvDomainMigratePrepare2",
                    "virDrvDomainMigratePrepare3",
                    "virDrvDomainMigratePrepare3Params",
                    "virDrvDomainMigratePrepareTunnel",
                    "virDrvDomainMigratePrepareTunnelParams",
                    "virDrvDomainMigratePrepareTunnel3",
                    "virDrvDomainMigratePrepareTunnel3Params",
                    "virDrvDomainMigratePerform",
                    "virDrvDomainMigratePerform3",
                    "virDrvDomainMigratePerform3Params",
                    "virDrvDomainMigrateConfirm",
                    "virDrvDomainMigrateConfirm3",
                    "virDrvDomainMigrateConfirm3Params",
                    "virDrvDomainMigrateBegin",
                    "virDrvDomainMigrateBegin3",
                    "virDrvDomainMigrateBegin3Params",
                    "virDrvDomainMigrateFinish",
                    "virDrvDomainMigrateFinish2",
                    "virDrvDomainMigrateFinish3",
                    "virDrvDomainMigrateFinish3Params",
                    "virDrvStreamInData",
                ]
                if drv in skip:
                    continue

                sym = drv.replace("virDrv", "vir")

                if sym not in symbols:
                    print("Driver method name %s doesn't match public API" %
                          drv)
                    status = 1
                continue

            m = re.search(r'''(\*vir\w+)\s*\)''', line)
            if m is not None:
                name = m.group(1)
                print("Bogus name %s" % name)
                status = 1
                continue

            m = re.search(r'''^\s*(virDrv\w+)\s+(\w+);\s*''', line)
            if m is not None:
                drv = m.group(1)
                field = m.group(2)

                tmp = drv.replace("virDrv", "")
                if tmp.startswith("NWFilter"):
                    tmp = "nwfilter" + tmp[8:]
                tmp = tmp[0:1].lower() + tmp[1:]

                if tmp != field:
                    print("Driver struct field %s should be named %s" %
                          (field, tmp))
                    status = 1

sys.exit(status)



#!/usr/bin/env python3
#
# Copyright (C) 2011-2019 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#
# Generate a set of systemtap functions for translating various
# RPC enum values into strings
#
#   python gensystemtap.py */*.x > libvirt_functions.stp
#

import re
import sys

funcs = {}

types = {}
status = {}
auth = {}


def load_file(fh):
    instatus = False
    intype = False
    inauth = False

    for line in fh:
        if re.search(r'''enum\s+virNetMessageType''', line):
            intype = True
        elif re.search(r'''enum\s+virNetMessageStatus''', line):
            instatus = True
        elif re.search(r'''enum remote_auth_type''', line):
            inauth = True
        elif "}" in line:
            intype = False
            instatus = False
            inauth = False
        elif instatus:
            m = re.search(r'''^\s+VIR_NET_(\w+)\s*=\s*(\d+),?$''', line)
            if m is not None:
                status[m.group(2)] = m.group(1).lower()
        elif intype:
            m = re.search(r'''^\s+VIR_NET_(\w+)\s*=\s*(\d+),?$''', line)
            if m is not None:
                types[m.group(2)] = m.group(1).lower()
        elif inauth:
            m = re.search(r'''^\s+REMOTE_AUTH_(\w+)\s*=\s*(\d+),?$''', line)
            if m is not None:
                auth[m.group(2)] = m.group(1).lower()
        else:
            m = re.search(r'''(?:VIR_)?(\w+?)(?:_PROTOCOL)?''' +
                          r'''_PROGRAM\s*=\s*0x([a-fA-F0-9]+)\s*;''', line)
            if m is not None:
                funcs[m.group(1).lower()] = {
                    "id": int(m.group(2), 16),
                    "version": None,
                    "progs": []
                }
                continue

            m = re.search(r'''(?:VIR_)?(\w+?)(?:_PROTOCOL)?_''' +
                          r'''(?:PROGRAM|PROTOCOL)_VERSION\s*=\s*(\d+)\s*;''',
                          line)
            if m is not None:
                funcs[m.group(1).lower()]["version"] = m.group(2)
                continue

            m = re.search(r'''(?:VIR_)?(\w+?)(?:_PROTOCOL)?''' +
                          r'''_PROC_(.*?)\s+=\s+(\d+)''', line)
            if m is not None:
                funcs[m.group(1).lower()]["progs"].insert(
                    int(m.group(3)), m.group(2).lower())


for file in sys.argv[1:]:
    with open(file, "r") as fh:
        load_file(fh)


def genfunc(name, varname, types):
    print("function %s(%s, verbose)" % (name, varname))
    print("{")

    first = True
    for typename in sorted(types.keys()):
        cond = "} else if"
        if first:
            cond = "if"
            first = False

        print("  %s (%s == %s) {" % (cond, varname, typename))
        print("      %sstr = \"%s\"" % (varname, types[typename]))

    print("  } else {")
    print("      %sstr = \"unknown\";" % varname)
    print("      verbose = 1;")
    print("  }")
    print("  if (verbose) {")
    print("      %sstr = %sstr . sprintf(\":%%d\", %s)" %
          (varname, varname, varname))
    print("  }")
    print("  return %sstr;" % varname)
    print("}")


genfunc("libvirt_rpc_auth_name", "type", auth)
genfunc("libvirt_rpc_type_name", "type", types)
genfunc("libvirt_rpc_status_name", "status", status)

print("function libvirt_rpc_program_name(program, verbose)")
print("{")

first = True
for funcname in sorted(funcs.keys()):
    cond = "} else if"
    if first:
        cond = "if"
        first = False

    print("  %s (program == %s) {" % (cond, funcs[funcname]["id"]))
    print("      programstr = \"%s\"" % funcname)

print("  } else {")
print("      programstr = \"unknown\";")
print("      verbose = 1;")
print("  }")
print("  if (verbose) {")
print("      programstr = programstr . sprintf(\":%d\", program)")
print("  }")
print("  return programstr;")
print("}")

print("function libvirt_rpc_procedure_name(program, version, proc, verbose)")
print("{")

first = True
for prog in sorted(funcs.keys()):
    cond = "} else if"
    if first:
        cond = "if"
        first = False

    print("  %s (program == %s && version == %s) {" %
          (cond, funcs[prog]["id"], funcs[prog]["version"]))

    pfirst = True
    for id in range(len(funcs[prog]["progs"])):
        pcond = "} else if"
        if pfirst:
            pcond = "if"
            pfirst = False

        print("      %s (proc == %s) {" % (pcond, id + 1))
        print("          procstr = \"%s\";" % funcs[prog]["progs"][id])

    print("      } else {")
    print("          procstr = \"unknown\";")
    print("          verbose = 1;")
    print("      }")

print("  } else {")
print("      procstr = \"unknown\";")
print("      verbose = 1;")
print("  }")
print("  if (verbose) {")
print("      procstr = procstr . sprintf(\":%d\", proc)")
print("  }")
print("  return procstr;")
print("}")



#!/usr/bin/env python3
#
# Copyright (C) 2017-2019 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.

import re
import sys

noninlined = {}
mocked = {}

# Functions in public header don't get the noinline annotation
noninlined["virEventAddTimeout"] = True
# This one confuses the script as its defined in the mock file
# but is actually just a local helper
noninlined["virMockStatRedirect"] = True


def scan_annotations(filename):
    with open(filename, "r") as fh:
        func = None
        for line in fh:
            m = re.search(r'''^\s*(\w+)\(''', line)
            if m is None:
                m = re.search(r'''^(?:\w+\*?\s+)+(?:\*\s*)?(\w+)\(''', line)
            if m is not None:
                name = m.group(1)
                if "ATTRIBUTE" not in name and "G_GNUC_" not in name:
                    func = name
            elif line.isspace():
                func = None

            if "G_GNUC_NO_INLINE" in line:
                if func is not None:
                    noninlined[func] = True


def scan_overrides(filename):
    with open(filename, "r") as fh:
        lineno = 0
        for line in fh:
            lineno = lineno + 1

            m = re.search(r'''^(\w+)\(''', line)
            if m is None:
                m = re.search(r'''^\w+\s*(?:\*\s*)?(\w+)\(''', line)
            if m is not None:
                name = m.group(1)
                if name.startswith("vir") or name.startswith("qemu") or name.startswith("libxl"):
                    mocked[name] = "%s:%d" % (filename, lineno)


for filename in sys.argv[1:]:
    if filename.endswith(".h"):
        scan_annotations(filename)
    elif filename.endswith("mock.c"):
        scan_overrides(filename)

warned = False
for func in mocked.keys():
    if func not in noninlined:
        warned = True
        print("%s is mocked at %s but missing noinline annotation" %
              (func, mocked[func]), file=sys.stderr)

if warned:
    sys.exit(1)
sys.exit(0)



#!/usr/bin/env python3

#
# hyperv_wmi_generator.py: generates most of the WMI type mapping code
#
# Copyright (C) 2011 Matthias Bolte <matthias.bolte@googlemail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#

import sys
import os
import os.path

separator = "/*" + ("*" * 50) + "*\n"
wmi_classes_by_name = {}


class WmiClass:
    """Represents WMI class and provides methods to generate C code."""

    def __init__(self, name, properties, uri_info):
        self.name = name
        self.properties = properties
        self.uri_info = uri_info

    def generate_classes_header(self):
        """Generate C header code and return it as string

        Declares:
          <class_name>_Data - used as hypervObject->data
          <class_name>_TypeInfo - used as wsman XmlSerializerInfo
          <class_name> - "inherits" hypervObject struct
        """

        name_upper = self.name.upper()

        header = separator
        header += " * %s\n" % self.name
        header += " */\n"
        header += "\n"
        header += "#define %s_WQL_SELECT \\\n" % name_upper
        header += "    \"SELECT * FROM %s \"\n" % self.name
        header += "\n"
        header += "extern hypervWmiClassInfoPtr %s_WmiInfo;\n\n" % self.name

        header += self._declare_data_structs()
        header += self._declare_hypervObject_struct()

        return header

    def generate_classes_source(self):
        """Returns a C code string defining wsman data structs

        Defines:
          <class_name>_Data struct
          <class_name>_WmiInfo - list holding metadata (e.g. request URIs) for the WMI class
        """

        source = separator
        source += " * %s\n" % self.name
        source += " */\n"

        source += "SER_START_ITEMS(%s_Data)\n" % self.name

        for property in self.properties:
            source += property.generate_classes_source(self.name)

        source += "SER_END_ITEMS(%s_Data);\n\n" % self.name

        # also generate typemap data while we're here
        source += "hypervCimType %s_Typemap[] = {\n" % self.name

        for property in self.properties:
            source += property.generate_typemap()
        source += '    { "", "", 0 },\n'  # null terminated
        source += '};\n\n'

        source += self._define_WmiInfo_struct()
        source += "\n\n"

        return source

    def generate_classes_typedef(self):
        """Returns C string for typedefs"""

        typedef = "typedef struct _%s %s;\n" % (self.name, self.name)
        typedef += "typedef struct _%s_Data %s_Data;\n" % (self.name, self.name)

        return typedef

    def _declare_data_structs(self):
        """Returns string C code declaring data structs.

        The *_Data structs are used as hypervObject->data. Each one has
        corresponding *_TypeInfo that is used for wsman unserialization of
        response XML into the *_Data structs.
        """

        header = "#define %s_RESOURCE_URI \\\n" % self.name.upper()
        header += "    \"%s\"\n" % self.uri_info.resourceUri
        header += "\n"
        header += "struct _%s_Data {\n" % self.name
        for property in self.properties:
            header += property.generate_classes_header()
        header += "};\n\n"
        header += "SER_DECLARE_TYPE(%s_Data);\n" % self.name

        return header

    def _declare_hypervObject_struct(self):
        """Return string for C code declaring hypervObject instance"""

        header = "\n/* must match hypervObject */\n"
        header += "struct _%s {\n" % self.name
        header += "    %s_Data *data;\n" % self.name
        header += "    hypervWmiClassInfoPtr info;\n"
        header += "    %s *next;\n" % self.name
        header += "};\n"

        header += "\n\n\n"

        return header

    def _define_WmiInfo_struct(self):
        """Return string for C code defining *_WmiInfo struct

        This struct holds info with meta-data needed to make wsman requests for the WMI class.
        """

        source = "hypervWmiClassInfoPtr %s_WmiInfo = &(hypervWmiClassInfo) {\n" % self.name
        source += "    .name = \"%s\",\n" % self.name
        source += "    .rootUri = %s,\n" % self.uri_info.rootUri
        source += "    .resourceUri = %s_RESOURCE_URI,\n" % self.name.upper()
        source += "    .serializerInfo = %s_Data_TypeInfo,\n" % self.name
        source += "    .propertyInfo = %s_Typemap\n" % self.name
        source += "};\n"

        return source


class ClassUriInfo:
    """Prepares URI information needed for wsman requests."""

    def __init__(self, wmi_name):
        if wmi_name.startswith("Msvm_"):
            self.rootUri = "ROOT_VIRTUALIZATION_V2"
            baseUri = "http://schemas.microsoft.com/wbem/wsman/1/wmi/root/virtualization/v2"
        else:
            self.rootUri = "ROOT_CIMV2"
            baseUri = "http://schemas.microsoft.com/wbem/wsman/1/wmi/root/cimv2"

        self.resourceUri = "%s/%s" % (baseUri, wmi_name)


class Property:
    typemap = {
        "boolean": "BOOL",
        "string": "STR",
        "datetime": "STR",
        "int8": "INT8",
        "sint8": "INT8",
        "int16": "INT16",
        "sint16": "INT16",
        "int32": "INT32",
        "sint32": "INT32",
        "int64": "INT64",
        "sint64": "INT64",
        "uint8": "UINT8",
        "uint16": "UINT16",
        "uint32": "UINT32",
        "uint64": "UINT64"
    }

    def __init__(self, type, name, is_array):
        if type not in Property.typemap:
            report_error("unhandled property type %s" % type)

        self.type = type
        self.name = name
        self.is_array = is_array

    def generate_classes_header(self):
        if self.is_array:
            return "    XML_TYPE_DYN_ARRAY %s;\n" % self.name
        else:
            return "    XML_TYPE_%s %s;\n" \
                   % (Property.typemap[self.type], self.name)

    def generate_classes_source(self, class_name):
        if self.is_array:
            return "    SER_NS_DYN_ARRAY(%s_RESOURCE_URI, \"%s\", 0, 0, %s),\n" \
                   % (class_name.upper(), self.name, self.type)
        else:
            return "    SER_NS_%s(%s_RESOURCE_URI, \"%s\", 1),\n" \
                   % (Property.typemap[self.type], class_name.upper(), self.name)

    def generate_typemap(self):
        return '    { "%s", "%s", %s },\n' % (self.name, self.type.lower(), str(self.is_array).lower())


def open_file(filename):
    return open(filename, "wt")


def report_error(message):
    print("error: " + message)
    sys.exit(1)


def parse_class(block, number):
    # expected format: class <name>
    header_items = block[0][1].split()

    if len(header_items) != 2:
        report_error("line %d: invalid block header" % (number))

    assert header_items[0] == "class"

    name = header_items[1]

    if name in wmi_classes_by_name:
        report_error("class '%s' has already been defined" % name)

    properties = []

    for line in block[1:]:
        # expected format: <type> <name>
        items = line[1].split()

        if len(items) != 2:
            report_error("line %d: invalid property" % line[0])

        if items[1].endswith("[]"):
            items[1] = items[1][:-2]
            is_array = True
        else:
            is_array = False

        properties.append(Property(type=items[0], name=items[1], is_array=is_array))

    wmi_classes_by_name[name] = WmiClass(name, properties, ClassUriInfo(name))


def main():
    if len(sys.argv) != 3:
        report_error("usage: %s srcdir builddir" % sys.argv[0])

    input_filename = os.path.join(sys.argv[1], "hyperv", "hyperv_wmi_generator.input")
    output_dirname = os.path.join(sys.argv[2], "hyperv")

    classes_typedef = open_file(os.path.join(output_dirname, "hyperv_wmi_classes.generated.typedef"))
    classes_header = open_file(os.path.join(output_dirname, "hyperv_wmi_classes.generated.h"))
    classes_source = open_file(os.path.join(output_dirname, "hyperv_wmi_classes.generated.c"))

    # parse input file
    number = 0
    block = None

    for line in open(input_filename, "rt").readlines():
        number += 1

        if "#" in line:
            line = line[:line.index("#")]

        line = line.lstrip().rstrip()

        if len(line) < 1:
            continue

        if line.startswith("class"):
            if block is not None:
                report_error("line %d: nested block found" % (number))
            else:
                block = []

        if block is not None:
            if line == "end":
                if block[0][1].startswith("class"):
                    parse_class(block, number)

                block = None
            else:
                block.append((number, line))

    # write output files
    notice = "/* Generated by hyperv_wmi_generator.py */\n\n\n\n"

    classes_typedef.write(notice)
    classes_header.write(notice)
    classes_source.write(notice)

    names = sorted(wmi_classes_by_name.keys())

    for name in names:
        cls = wmi_classes_by_name[name]

        classes_typedef.write(cls.generate_classes_typedef())
        classes_header.write(cls.generate_classes_header())
        classes_source.write(cls.generate_classes_source())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#

import re
import sys


def checkdriverimpls(filename):
    intable = False
    mainprefix = None

    errs = False
    with open(filename, "r") as fh:
        lineno = 0
        for line in fh:
            lineno = lineno + 1
            if intable:
                if "}" in line:
                    intable = False
                    mainprefix = None
                    continue

                m = re.search(r'''\.(\w+)\s*=\s*(\w+),?''', line)
                if m is not None:
                    api = m.group(1)
                    impl = m.group(2)

                    if api in ["no", "name"]:
                        continue
                    if impl in ["NULL"]:
                        continue

                    suffix = impl
                    prefix = re.sub(r'''^([a-z]+)(.*?)$''', r'''\1''', impl)

                    if mainprefix is not None:
                        if mainprefix != prefix:
                            print(("%s:%d Bad prefix '%s' for API '%s', " +
                                   "expecting '%s'") %
                                  (filename, lineno, prefix, api, mainprefix),
                                  file=sys.stderr)
                            errs = True
                    else:
                        mainprefix = prefix

                    if not api.startswith(mainprefix):
                        suffix = re.sub(r'''^[a-z]+''', "", suffix)
                        suffix = re.sub(r'''^([A-Z]+)''',
                                        lambda m: m.group(1).lower(), suffix)

                    if api != suffix:
                        want = api
                        if want.startswith("nwf"):
                            want = "NWF" + want[3:]

                        if not api.startswith(mainprefix):
                            want = re.sub(r'''^([a-z])''',
                                          lambda m: m.group(1).upper(), want)
                            want = mainprefix + want

                        print(("%s:%d Bad impl name '%s' for API " +
                               "'%s', expecting '%s'") %
                              (filename, lineno, impl, api, want),
                              file=sys.stderr)
                        errs = True
            else:
                m = re.search(r'''^(?:static\s+)?(vir(?:\w+)?Driver)''' +
                              r'''\s+(?!.*;)''', line)
                if m is not None:
                    drv = m.group(1)
                    if drv in [
                            "virNWFilterCallbackDriver",
                            "virNWFilterTechDriver",
                            "virConnectDriver"]:
                        continue
                    intable = True

    return errs


status = 0
for filename in sys.argv[1:]:
    if checkdriverimpls(filename):
        status = 1
sys.exit(status)



#!/usr/bin/env python3
#
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see
# <http://www.gnu.org/licenses/>.
#

import re
import sys

objects = [
    "CONNECT", "DOMAIN", "INTERFACE",
    "NETWORK_PORT", "NETWORK", "NODE_DEVICE",
    "NWFILTER_BINDING", "NWFILTER",
    "SECRET", "STORAGE_POOL", "STORAGE_VOL",
]


classes = {}

for obj in objects:
    klass = obj.lower()

    klass = re.sub(r'''(^\w|_\w)''', lambda a: a.group(1).upper(), klass)
    klass = klass.replace("_", "")
    klass = klass.replace("Nwfilter", "NWFilter")
    klass = "vir" + klass + "Ptr"

    classes[obj] = klass


objectstr = "|".join(objects)

opts = {}
in_opts = {}

perms = {}

aclfile = sys.argv[1]
with open(aclfile, "r") as fh:
    for line in fh:
        if in_opts:
            if line.find("*/") != -1:
                in_opts = False
            else:
                m = re.search(r'''\*\s*\@(\w+):\s*(.*?)\s*$''', line)
                if m is not None:
                    opts[m.group(1)] = m.group(2)
        elif line.find("**") != -1:
            in_opts = True
        else:
            m = re.search(r'''VIR_ACCESS_PERM_(%s)_((?:\w|_)+),''' %
                          objectstr, line)
            if m is not None:
                obj = m.group(1)
                perm = m.group(2).lower()
                if perm == "last":
                    continue

                perm = perm.replace("_", "-")

                if obj not in perms:
                    perms[obj] = {}
                perms[obj][perm] = {
                    "desc": opts.get("desc", None),
                    "message": opts.get("message", None),
                    "anonymous": opts.get("anonymous", None),
                }
                opts = {}

print('<?xml version="1.0" encoding="UTF-8"?>')
print('<!DOCTYPE html>')
print('<html xmlns="http://www.w3.org/1999/xhtml">')
print('  <body>')

for obj in sorted(perms.keys()):
    klass = classes[obj]

    olink = "object_" + obj.lower()

    print('    <h3><a id="%s">%s</a></h3>' % (olink, klass))
    print('    <table>')
    print('      <thead>')
    print('        <tr>')
    print('          <th>Permission</th>')
    print('          <th>Description</th>')
    print('        </tr>')
    print('      </thead>')
    print('      <tbody>')

    for perm in sorted(perms[obj].keys()):
        description = perms[obj][perm]["desc"]

        if description is None:
            raise Exception("missing description for %s.%s" % (obj, perm))

        plink = "perm_" + obj.lower() + "_" + perm.lower()
        plink = plink.replace("-", "_")

        print('        <tr>')
        print('          <td><a id="%s">%s</a></td>' % (plink, perm))
        print('          <td>%s</td>' % description)
        print('        </tr>')

    print('      </tbody>')
    print('    </table>')

print('  </body>')
print('</html>')



#!/usr/bin/env python3

import argparse
import fcntl
import json
import os
import re
import struct
import subprocess
import sys
import xml.etree.ElementTree


_KEYS = {
    "cpuid": ["eax_in", "ecx_in"],
    "msr": ["index"],
}

_REGS = {
    "cpuid": ["eax", "ebx", "ecx", "edx"],
    "msr": ["eax", "edx"],
}


def gather_name(args):
    if args.name:
        return args.name

    with open("/proc/cpuinfo", "rt") as f:
        for line in f.readlines():
            if line.startswith("model name"):
                return line.split(":", 2)[1].strip()

    exit("Error: '/proc/cpuinfo' does not contain a model name.\n"
         "Use '--model' to set a model name.")


def gather_cpuid_leaves(args):
    def mask(regs, eax_in, ecx_in, eax_mask, ebx_mask, ecx_mask, edx_mask):
        if regs["eax_in"] == eax_in and regs["ecx_in"] == ecx_in:
            regs["eax"] &= eax_mask
            regs["ebx"] &= ebx_mask
            regs["ecx"] &= ecx_mask
            regs["edx"] &= edx_mask

    leave_pattern = re.compile(
        "^\\s*"
        "(0x[0-9a-f]+)\\s*"
        "(0x[0-9a-f]+):\\s*"
        "eax=(0x[0-9a-f]+)\\s*"
        "ebx=(0x[0-9a-f]+)\\s*"
        "ecx=(0x[0-9a-f]+)\\s*"
        "edx=(0x[0-9a-f]+)\\s*$")

    cpuid = args.path_to_cpuid or "cpuid"
    try:
        output = subprocess.check_output(
            [cpuid, "-1r"],
            universal_newlines=True)
    except FileNotFoundError as e:
        exit("Error: '{}' not found.\n'cpuid' can be usually found in a "
             "package named identically. If your distro does not provide such "
             "package, you can find the sources or binary packages at "
             "'http://www.etallen.com/cpuid.html'.".format(e.filename))

    for line in output.split("\n"):
        match = leave_pattern.match(line)
        if not match:
            continue
        regs = {
            "eax_in": int(match.group(1), 0),
            "ecx_in": int(match.group(2), 0),
            "eax": int(match.group(3), 0),
            "ebx": int(match.group(4), 0),
            "ecx": int(match.group(5), 0),
            "edx": int(match.group(6), 0)}

        # local apic id. Pretend to always run on logical processor #0.
        mask(regs, 0x01, 0x00, 0xffffffff, 0x00ffffff, 0xffffffff, 0xffffffff)
        mask(regs, 0x0b, 0x00, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffff00)
        mask(regs, 0x0b, 0x01, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffff00)

        yield regs


def gather_msr():
    msrs = dict()
    addresses = [
        0x10a,  # IA32_ARCH_CAPABILITIES_MSR
        0xcf,   # IA32_CORE_CAPABILITY_MSR
    ]
    KVM_GET_MSRS = 0xc008ae88

    try:
        with open("/dev/cpu/0/msr", "rb") as f:
            for addr in addresses:
                f.seek(addr)
                buf = f.read(8)
                msrs[addr] = struct.unpack("=Q", buf)[0]
            return "", msrs
    except IOError as e:
        print("Warning: {}".format(e), file=sys.stderr)

    try:
        with open("/dev/kvm", "rb") as f:
            for addr in addresses:
                bufIn = struct.pack("=LLLLQ", 1, 0, addr, 0, 0)
                bufOut = fcntl.ioctl(f, KVM_GET_MSRS, bufIn)
                msrs[addr] = struct.unpack("=LLLLQ", bufOut)[4]
            return " via KVM", msrs
    except IOError as e:
        print("Warning: {}".format(e), file=sys.stderr)

    return None, {}


def call_qemu(qemu, qmp_cmds):
    cmd = [
        qemu,
        "-machine", "accel=kvm",
        "-cpu", "host",
        "-nodefaults",
        "-nographic",
        "-qmp", "stdio"]

    stdin = list()
    stdin.append("{\"execute\": \"qmp_capabilities\"}")
    stdin.extend([json.dumps(o) for o in qmp_cmds])
    stdin.append("{\"execute\": \"quit\"}")

    try:
        output = subprocess.check_output(
            cmd,
            universal_newlines=True,
            input="\n".join(stdin))
    except subprocess.CalledProcessError:
        exit("Error: Non-zero exit code from '{}'.".format(qemu))
    except FileNotFoundError:
        exit("Error: File not found: '{}'.".format(qemu))

    for line in output.split("\n"):
        if not line:
            continue
        response = json.loads(line)
        if "return" in response and not response["return"]:
            continue
        if response.get("event") == "SHUTDOWN":
            continue
        yield response


def gather_model(args):
    output = call_qemu(args.path_to_qemu, [
        {
            "execute": "query-cpu-model-expansion",
            "arguments":
            {
                "type": "static",
                "model": {"name": "host"}
            },
            "id": "model-expansion"
        }])

    static_model = None
    for o in output:
        if o.get("id") == "model-expansion":
            static_model = o["return"]["model"]

    if static_model:
        return call_qemu(args.path_to_qemu, [
            {
                "execute": "query-cpu-model-expansion",
                "arguments":
                {
                    "type": "full",
                    "model": static_model
                },
                "id": "model-expansion"
            },
            {
                "execute": "query-cpu-definitions",
                "id": "definitions"
            }
        ])
    else:
        return call_qemu(args.path_to_qemu, [
            {
                "execute": "qom-get",
                "arguments":
                {
                    "path": "/machine/unattached/device[0]",
                    "property": "feature-words"
                },
                "id": "feature-words"
            },
            {
                "execute": "qom-get",
                "arguments":
                {
                    "path": "/machine/unattached/device[0]",
                    "property": "family"
                },
                "id": "family"
            },
            {
                "execute": "qom-get",
                "arguments":
                {
                    "path": "/machine/unattached/device[0]",
                    "property": "model"
                },
                "id": "model"
            },
            {
                "execute": "qom-get",
                "arguments":
                {
                    "path": "/machine/unattached/device[0]",
                    "property": "stepping"
                },
                "id": "stepping"
            },
            {
                "execute": "qom-get",
                "arguments":
                {
                    "path": "/machine/unattached/device[0]",
                    "property": "model-id"
                },
                "id": "model-id"
            },
            {
                "execute": "query-cpu-definitions",
                "id": "definitions"
            }
        ])


def gather(args):
    result = dict()
    result["name"] = gather_name(args)
    result["leaves"] = list(gather_cpuid_leaves(args))
    result["via"], result["msr"] = gather_msr()
    result["model"] = list(gather_model(args))
    return result


def parse_filename(data):
    filename = data["name"].strip()
    filename = re.sub("[ -]+ +", " ", filename)
    filename = re.sub("\\(([Rr]|[Tt][Mm])\\)", "", filename)
    filename = re.sub(".*(Intel|AMD) ", "", filename)
    filename = re.sub(" (Duo|Quad|II X[0-9]+)", " ", filename)
    filename = re.sub(" (CPU|[Pp]rocessor)", "", filename)
    filename = re.sub(" @.*", "", filename)
    filename = re.sub(" APU .*", "", filename)
    filename = re.sub(" SE$", "", filename)
    filename = re.sub(" ", "-", filename)
    return "x86_64-cpuid-{}".format(filename)


def output_xml(data, filename):
    leave_template = \
        "  <cpuid" \
        " eax_in='0x{0[eax_in]:08x}'" \
        " ecx_in='0x{0[ecx_in]:02x}'" \
        " eax='0x{0[eax]:08x}'" \
        " ebx='0x{0[ebx]:08x}'" \
        " ecx='0x{0[ecx]:08x}'" \
        " edx='0x{0[edx]:08x}'" \
        "/>\n"

    msr_template = "  <msr index='0x{:x}' edx='0x{:08x}' eax='0x{:08x}'/>\n"

    print(filename)
    with open(filename, "wt") as f:
        f.write("<!-- {} -->\n".format(data["name"]))
        f.write("<cpudata arch='x86'>\n")
        for leave in data["leaves"]:
            f.write(leave_template.format(leave))
        for key, value in sorted(data["msr"].items()):
            f.write(msr_template.format(
                int(key),
                0xffffffff & (value >> 32),
                0xffffffff & (value >> 0)))
        f.write("</cpudata>\n")


def output_json(data, filename):
    replies = list()
    for reply in data["model"]:
        if "QMP" in reply:
            continue
        if "timestamp" in reply:
            continue
        if "return" in reply and not reply["return"]:
            continue
        replies.append(reply)

    if not replies:
        return

    if "model-expansion" not in [reply.get("id") for reply in replies]:
        exit(
            "Error: Missing query-cpu-model-expansion reply in "
            "{}".format(filename))

    print(filename)
    with open(filename, "wt") as f:
        for reply in replies:
            if reply is not replies[0]:
                f.write("\n")
            json.dump(reply, f, indent=2)
            f.write("\n")


def parse(args, data):
    filename = parse_filename(data)
    filename_xml = "{}.xml".format(filename)
    filename_json = "{}.json".format(filename)

    output_xml(data, filename_xml)
    output_json(data, filename_json)

    if not os.path.isfile(filename_json):
        return
    if os.path.getsize(filename_json) == 0:
        return

    args.json_files = getattr(args, "json_files", list()) + [filename_json]


def checkFeature(cpuData, feature):
    for key in ["type"] + _KEYS.get(feature["type"], list()):
        if feature[key] not in cpuData:
            return False
        cpuData = cpuData[feature[key]]

    for reg in _REGS.get(feature["type"], list()):
        if feature[reg] > 0 and feature[reg] == feature[reg] & cpuData[reg]:
            return True
    return False


def addFeature(cpuData, feature):
    for key in ["type"] + _KEYS.get(feature["type"], list()):
        if feature[key] not in cpuData:
            cpuData[feature[key]] = dict()
        cpuData = cpuData[feature[key]]

    for reg in _REGS.get(feature["type"], list()):
        cpuData[reg] = cpuData.get(reg, 0) | feature[reg]


def parseQemu(path, features):
    cpuData = {}
    with open(path, "r") as f:
        data, pos = json.JSONDecoder().raw_decode(f.read())

    for (prop, val) in data["return"]["model"]["props"].items():
        if val and prop in features:
            addFeature(cpuData, features[prop])

    return cpuData


def parseCPUData(path):
    cpuData = dict()
    for f in xml.etree.ElementTree.parse(path).getroot():
        if f.tag not in ("cpuid", "msr"):
            continue

        feature = {"type": f.tag}
        for reg in _KEYS[f.tag] + _REGS[f.tag]:
            feature[reg] = int(f.attrib.get(reg, "0"), 0)
        addFeature(cpuData, feature)
    return cpuData


def parseMap():
    path = os.path.dirname(sys.argv[0])
    path = os.path.join(path, "..", "..", "src", "cpu_map", "x86_features.xml")

    cpuMap = dict()
    for f in xml.etree.ElementTree.parse(path).getroot().iter("feature"):
        if f[0].tag not in ("cpuid", "msr"):
            continue

        feature = {"type": f[0].tag}
        for reg in _KEYS[f[0].tag] + _REGS[f[0].tag]:
            feature[reg] = int(f[0].attrib.get(reg, "0"), 0)
        cpuMap[f.attrib["name"]] = feature
    return cpuMap


def formatCPUData(cpuData, path, comment):
    print(path)
    with open(path, "w") as f:
        f.write("<!-- " + comment + " -->\n")
        f.write("<cpudata arch='x86'>\n")

        cpuid = cpuData["cpuid"]
        for eax_in in sorted(cpuid.keys()):
            for ecx_in in sorted(cpuid[eax_in].keys()):
                leaf = cpuid[eax_in][ecx_in]
                line = ("  <cpuid eax_in='0x%08x' ecx_in='0x%02x' "
                        "eax='0x%08x' ebx='0x%08x' "
                        "ecx='0x%08x' edx='0x%08x'/>\n")
                f.write(line % (
                        eax_in, ecx_in,
                        leaf["eax"], leaf["ebx"], leaf["ecx"], leaf["edx"]))

        if "msr" in cpuData:
            msr = cpuData["msr"]
            for index in sorted(msr.keys()):
                f.write("  <msr index='0x%x' edx='0x%08x' eax='0x%08x'/>\n" %
                        (index, msr[index]['edx'], msr[index]['eax']))

        f.write("</cpudata>\n")


def diff(args):
    cpuMap = parseMap()

    for jsonFile in args.json_files:
        cpuDataFile = jsonFile.replace(".json", ".xml")
        enabledFile = jsonFile.replace(".json", "-enabled.xml")
        disabledFile = jsonFile.replace(".json", "-disabled.xml")

        cpuData = parseCPUData(cpuDataFile)
        qemu = parseQemu(jsonFile, cpuMap)

        enabled = dict()
        disabled = dict()
        for feature in cpuMap.values():
            if checkFeature(qemu, feature):
                addFeature(enabled, feature)
            elif checkFeature(cpuData, feature):
                addFeature(disabled, feature)

        formatCPUData(enabled, enabledFile, "Features enabled by QEMU")
        formatCPUData(disabled, disabledFile, "Features disabled by QEMU")


def main():
    parser = argparse.ArgumentParser(description="Gather cpu test data")
    parser.add_argument(
        "--name",
        help="CPU model name. "
        "If unset, model name is read from '/proc/cpuinfo'.")
    parser.add_argument(
        "--path-to-cpuid",
        metavar="PATH",
        help="Path to 'cpuid' utility. "
        "If unset, the first executable 'cpuid' in $PATH is used.")
    parser.add_argument(
        "--path-to-qemu",
        metavar="PATH",
        help="Path to qemu. "
        "If unset, will try '/usr/bin/qemu-system-x86_64', "
        "'/usr/bin/qemu-kvm', and '/usr/libexec/qemu-kvm'.")
    subparsers = parser.add_subparsers(dest="action")
    subparsers.add_parser(
        "gather",
        help="Acquire data on target system and outputs to stdout. "
        "This is the default. ")
    subparsers.add_parser(
        "parse",
        help="Reads data from stdin and parses data for libvirt use.")
    subparsers.add_parser(
        "full",
        help="Equivalent to `cpu-data.py gather | cpu-data.py parse`.")
    diffparser = subparsers.add_parser(
        "diff",
        help="Diff json description of CPU model against known features.")
    diffparser.add_argument(
        "json_files",
        nargs="+",
        metavar="FILE",
        type=os.path.realpath,
        help="Path to one or more json CPU model descriptions.")

    args = parser.parse_args()

    if not args.action:
        args.action = "gather"

    if not args.path_to_qemu:
        args.path_to_qemu = "qemu-system-x86_64"
        search = [
            "/usr/bin/qemu-system-x86_64",
            "/usr/bin/qemu-kvm",
            "/usr/libexec/qemu-kvm"]
        for f in search:
            if os.path.isfile(f):
                args.path_to_qemu = f

    if args.action in ["gather", "full"]:
        data = gather(args)
        if args.action == "gather":
            json.dump(data, sys.stdout, indent=2)

    if args.action in ["parse", "full"]:
        if args.action == "parse":
            data = json.load(sys.stdin)
        parse(args, data)

    if "json_files" in args:
        diff(args)


if __name__ == "__main__":
    main()
